from typing import Optional, Callable, Awaitable

from aiohttp import web


class Middleware:
    """
    Middleware functions are called before every routes registered by dazzler.

    Add callback by returning a function taking the response as argument.

    .. code-block:: python3

        from dazzler.system import Middleware

        class MyMiddleware(Middleware):
            async def __call__(self, request):

                async def set_cookie(response):
                    response.set_cookie('cookie', 'value')

                return set_cookie
    """
    async def __call__(
            self,
            request: web.Request,
    ) -> Optional[Callable[[web.Response], Awaitable]]:
        raise NotImplementedError



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class RadioList(Component):
    """
    A radio button list where only a single value can be selected.
    """
    _aspect_keys = []
    _children = []
    options = Aspect(required=True, docstring="Items in the radio list with labels, values and elements attributes.")  # noqa: E501
    value = Aspect(docstring="Selected radio button.")
    style = Aspect(docstring="Style object of the container.")
    class_name = Aspect(docstring="Class name of the container.")
    labels_class_name = Aspect(docstring="Global class name to give to labels.")  # noqa: E501
    labels_style = Aspect(docstring="Global labels style objects.")
    options_class_name = Aspect(docstring="Global options class name.")
    options_style = Aspect(docstring="Global style object of options.")
    id = Aspect(docstring="Html id of the component, otherwise the identity is used.")  # noqa: E501

    def __init__(
            self,
            options: typing.List[typing.Dict[str, typing.Union[str, typing.Union[str, typing.Union[float, int]]]]],  # noqa: E501
            value: typing.Optional[typing.Union[str, typing.Union[float, int]]] = UNDEFINED,  # noqa: E501
            style: typing.Optional[typing.Dict] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            labels_class_name: typing.Optional[str] = UNDEFINED,
            labels_style: typing.Optional[typing.Dict] = UNDEFINED,
            options_class_name: typing.Optional[str] = UNDEFINED,
            options_style: typing.Optional[typing.Dict] = UNDEFINED,
            id: typing.Optional[str] = UNDEFINED,
            identity: str = None
    ):
        """
        :param options: Items in the radio list with labels, values and
            elements attributes.
        :param value: Selected radio button.
        :param style: Style object of the container.
        :param class_name: Class name of the container.
        :param labels_class_name: Global class name to give to labels.
        :param labels_style: Global labels style objects.
        :param options_class_name: Global options class name.
        :param options_style: Global style object of options.
        :param id: Html id of the component, otherwise the identity is
            used.
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class TreeView(Component):
    """
    A tree of nested items.

    :CSS:

        ``dazzler-extra-tree-view``
        - ``tree-item``
        - ``tree-item-label``
        - ``tree-sub-items``
        - ``tree-caret``
        - ``selected``
        - ``level-{n}``

    :example:

    .. literalinclude:: ../../tests/components/pages/treeview.py
    """
    _aspect_keys = []
    _children = []
    items = Aspect(required=True, docstring="An array of items to render recursively.")  # noqa: E501
    selected = Aspect(docstring="Last clicked path identifier joined by dot.")
    expanded_items = Aspect(default=UNDEFINED, docstring="Identifiers that have sub items and are open.READONLY. (default=[])")  # noqa: E501
    nest_icon_collapsed = Aspect(default='⏵', docstring="Icon to show when sub items are hidden. (default='⏵')")  # noqa: E501
    nest_icon_expanded = Aspect(default='⏷', docstring="Icon to show when sub items are shown. (default='⏷')")  # noqa: E501
    class_name = Aspect()
    style = Aspect()

    def __init__(
            self,
            items: typing.List[typing.Union[str, typing.Dict[str, typing.Union[str, typing.List[typing.Any]]]]],  # noqa: E501
            selected: typing.Optional[str] = UNDEFINED,
            expanded_items: typing.Optional[typing.List] = UNDEFINED,
            nest_icon_collapsed: typing.Optional[str] = '⏵',
            nest_icon_expanded: typing.Optional[str] = '⏷',
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            identity: str = None
    ):
        """
        :param items: An array of items to render recursively.
        :param selected: Last clicked path identifier joined by dot.
        :param expanded_items: Identifiers that have sub items and are
            open. READONLY. (default=[])
        :param nest_icon_collapsed: Icon to show when sub items are hidden.
            (default='⏵')
        :param nest_icon_expanded: Icon to show when sub items are shown.
            (default='⏷')
        :param class_name:
        :param style:
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Drawer(Component):
    """
    Draw content from the sides of the screen.
    """
    _aspect_keys = []
    _children = []
    children = Aspect(children=True)
    opened = Aspect()
    style = Aspect()
    class_name = Aspect()
    side = Aspect(default="'top'", docstring="Side which open. (Possible values: 'top', 'left', 'right', 'bottom') (default="'top'")")  # noqa: E501

    def __init__(
            self,
            children: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            opened: typing.Optional[bool] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            side: typing.Optional[typing.Any] = "'top'",
            identity: str = None
    ):
        """
        :param children:
        :param opened:
        :param style:
        :param class_name:
        :param side: Side which open. (Possible values: 'top', 'left',
            'right', 'bottom') (default="'top'")
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class DataList(Component):
    """
    A html datalist (select with autocomplete).
    """
    _aspect_keys = []
    _children = []
    options = Aspect(required=True, docstring="Options of the datalist.")
    id = Aspect(docstring="Unique id for the component.")
    class_name = Aspect(docstring="CSS class.")
    value = Aspect(docstring="Value of the text input.")
    data_value = Aspect(docstring="The value of the selected option if found.")
    title = Aspect(docstring="Tooltip.")

    def __init__(
            self,
            options: typing.List[typing.Dict[str, typing.Union[typing.Any, str]]],  # noqa: E501
            id: typing.Optional[str] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            value: typing.Optional[str] = UNDEFINED,
            data_value: typing.Optional[typing.Any] = UNDEFINED,
            title: typing.Optional[str] = UNDEFINED,
            identity: str = None
    ):
        """
        :param options: Options of the datalist.
        :param id: Unique id for the component.
        :param class_name: CSS class.
        :param value: Value of the text input.
        :param data_value: The value of the selected option if found.
        :param title: Tooltip.
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Html(Component):
    """
    Html tag wrapper, give any props as ``attributes``.
    Listen to events with the readonly event aspect containing the
    latest event fired.
    """
    _aspect_keys = []
    _children = []
    tag = Aspect(required=True, docstring="Tag name of the component.")
    children = Aspect(children=True, docstring="Children of the html tag.")
    id = Aspect(docstring="Id of the element in DOM.")
    class_name = Aspect(docstring="Class of the element.")
    attributes = Aspect(docstring="Any other html attributes relevant to the html tag")  # noqa: E501
    events = Aspect(docstring="Events to subscribe.")
    event = Aspect(docstring="Last event fired.")

    def __init__(
            self,
            tag: str,
            children: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            id: typing.Optional[str] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            attributes: typing.Optional[typing.Dict] = UNDEFINED,
            events: typing.Optional[typing.List] = UNDEFINED,
            event: typing.Optional[typing.Dict] = UNDEFINED,
            identity: str = None
    ):
        """
        :param tag: Tag name of the component.
        :param children: Children of the html tag.
        :param id: Id of the element in DOM.
        :param class_name: Class of the element.
        :param attributes: Any other html attributes relevant to the html
            tag
        :param events: Events to subscribe.
        :param event: Last event fired.
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Container(Component):
    """
    Virtual div

    :CSS:

        ``dazzler-core-container``.
        - ``scroll``
        - ``flex``
        - ``row``
        - ``column``
        - ``btn``
    """
    _aspect_keys = []
    _children = []
    children = Aspect(children=True)
    id = Aspect()
    class_name = Aspect()
    style = Aspect()
    title = Aspect()
    draggable = Aspect()
    clicks = Aspect(default=0, docstring="Number of times the container was clicked. (default=0)")  # noqa: E501

    def __init__(
            self,
            children: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            id: typing.Optional[str] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            title: typing.Optional[str] = UNDEFINED,
            draggable: typing.Optional[bool] = UNDEFINED,
            clicks: typing.Optional[typing.Union[float, int]] = 0,
            identity: str = None
    ):
        """
        :param children:
        :param id:
        :param class_name:
        :param style:
        :param title:
        :param draggable:
        :param clicks: Number of times the container was clicked.
            (default=0)
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Login(Component):
    """
    A login form to include on dazzler pages.

    :CSS:

        ``dazzler-auth-login``
        - ``login-field``
        - ``login-label``
        - ``login-input``
        - ``login-username``
        - ``login-password``
        - ``login-button``
    """
    _aspect_keys = []
    _children = []
    login_url = Aspect(required=True, docstring="The url to perform login.")
    next_url = Aspect(docstring="Redirect to this page after login.")
    method = Aspect(default='POST', docstring="Method to submit the login form. (default='POST')")  # noqa: E501
    username_label = Aspect(default='Username', docstring="The label to show before the. (default='Username')")  # noqa: E501
    password_label = Aspect(default='Password', docstring="Label to replace the password. (default='Password')")  # noqa: E501
    submit_label = Aspect(default='Login', docstring="Label for the submit button. (default='Login')")  # noqa: E501
    horizontal = Aspect(docstring="Style the form with the fields side by side.")  # noqa: E501
    bordered = Aspect(docstring="Apply a border around the")
    placeholder_labels = Aspect(docstring="Put the label in placeholder attribute instead of a `<label>` element.")  # noqa: E501
    header = Aspect(children=True, docstring="Included as first child of the form.Wrapped under div with CSS class ``login-header``")  # noqa: E501
    footer = Aspect(children=True, docstring="Included at bottom of the login form.Wrapped under div with CSS class ``login-footer``")  # noqa: E501
    errors = Aspect(docstring="Form errors")
    class_name = Aspect()
    style = Aspect()

    def __init__(
            self,
            login_url: str,
            next_url: typing.Optional[str] = UNDEFINED,
            method: typing.Optional[str] = 'POST',
            username_label: typing.Optional[str] = 'Username',
            password_label: typing.Optional[str] = 'Password',
            submit_label: typing.Optional[str] = 'Login',
            horizontal: typing.Optional[bool] = UNDEFINED,
            bordered: typing.Optional[bool] = UNDEFINED,
            placeholder_labels: typing.Optional[bool] = UNDEFINED,
            header: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            footer: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            errors: typing.Optional[typing.Dict] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            identity: str = None
    ):
        """
        :param login_url: The url to perform login.
        :param next_url: Redirect to this page after login.
        :param method: Method to submit the login form. (default='POST')
        :param username_label: The label to show before the.
            (default='Username')
        :param password_label: Label to replace the password.
            (default='Password')
        :param submit_label: Label for the submit button. (default='Login')
        :param horizontal: Style the form with the fields side by side.
        :param bordered: Apply a border around the
        :param placeholder_labels: Put the label in placeholder attribute
            instead of a `<label>` element.
        :param header: Included as first child of the form. Wrapped under
            div with CSS class ``login-header``
        :param footer: Included at bottom of the login form. Wrapped under
            div with CSS class ``login-footer``
        :param errors: Form errors
        :param class_name:
        :param style:
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class TimePicker(Component):
    """
    Time input with fallback for ie and safari.

    :CSS:

        ``dazzler-calendar-time-picker``
        - ``time-input``
        - ``fallback-timepicker``
        - ``time-range``
        - ``time-value``
    """
    _aspect_keys = []
    _children = []
    value = Aspect(default='00:00', docstring="Time value formatted as HH:MM (default='00:00')")  # noqa: E501
    fallback_mode = Aspect(docstring="Use a custom picker instead of browser dependant.Automatically set when no time input support detected.")  # noqa: E501
    mode = Aspect(default="'24h'", docstring="What to display when in fallback mode. (Possible values: '24h', 'AM/PM') (default="'24h'")")  # noqa: E501

    def __init__(
            self,
            value: typing.Optional[str] = '00:00',
            fallback_mode: typing.Optional[bool] = UNDEFINED,
            mode: typing.Optional[typing.Any] = "'24h'",
            identity: str = None
    ):
        """
        :param value: Time value formatted as HH:MM (default='00:00')
        :param fallback_mode: Use a custom picker instead of browser
            dependant. Automatically set when no time input support detected.
        :param mode: What to display when in fallback mode. (Possible
            values: '24h', 'AM/PM') (default="'24h'")
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Notice(Component):
    """
    Browser notifications with permissions handling.
    """
    _aspect_keys = []
    _children = []
    permission = Aspect(docstring="Permission granted by the user (READONLY) (Possible values: 'denied', 'granted', 'default', 'unsupported')")  # noqa: E501
    title = Aspect(required=True)
    lang = Aspect(docstring="The notification's language, as specified using a DOMString representing a BCP 47 language tag.")  # noqa: E501
    body = Aspect(docstring="A DOMString representing the body text of the notification, which will be displayed below the title.")  # noqa: E501
    badge = Aspect(docstring="A USVString containing the URL of the image used to represent the notification when there is not enough space to display the notification itself.")  # noqa: E501
    tag = Aspect(docstring="A DOMString representing an identifying tag for the notification.")  # noqa: E501
    icon = Aspect(docstring="A USVString containing the URL of an icon to be displayed in the notification.")  # noqa: E501
    image = Aspect(docstring="a USVString containing the URL of an image to be displayed in the notification.")  # noqa: E501
    vibrate = Aspect(docstring="A vibration pattern for the device's vibration hardware to emit when the notification fires.")  # noqa: E501
    require_interaction = Aspect(default=False, docstring="Indicates that a notification should remain active until the user clicks or dismisses it, rather than closing automatically. The default value is false. (default=False)")  # noqa: E501
    displayed = Aspect(docstring="Set to true to display the notification.")
    clicks = Aspect(default=0, docstring="(default=0)")
    clicks_timestamp = Aspect(default=-1, docstring="(default=-1)")
    closes = Aspect(default=0, docstring="Number of times the notification was closed. (default=0)")  # noqa: E501
    closes_timestamp = Aspect(default=-1, docstring="(default=-1)")

    def __init__(
            self,
            title: str,
            permission: typing.Optional[typing.Any] = UNDEFINED,
            lang: typing.Optional[str] = UNDEFINED,
            body: typing.Optional[str] = UNDEFINED,
            badge: typing.Optional[str] = UNDEFINED,
            tag: typing.Optional[str] = UNDEFINED,
            icon: typing.Optional[str] = UNDEFINED,
            image: typing.Optional[str] = UNDEFINED,
            vibrate: typing.Optional[typing.Union[typing.Union[float, int], typing.List[typing.Union[float, int]]]] = UNDEFINED,  # noqa: E501
            require_interaction: typing.Optional[bool] = False,
            displayed: typing.Optional[bool] = UNDEFINED,
            clicks: typing.Optional[typing.Union[float, int]] = 0,
            clicks_timestamp: typing.Optional[typing.Union[float, int]] = -1,
            closes: typing.Optional[typing.Union[float, int]] = 0,
            closes_timestamp: typing.Optional[typing.Union[float, int]] = -1,
            identity: str = None
    ):
        """
        :param permission: Permission granted by the user (READONLY)
            (Possible values: 'denied', 'granted', 'default', 'unsupported')
        :param title:
        :param lang: The notification's language, as specified using a
            DOMString representing a BCP 47 language tag.
        :param body: A DOMString representing the body text of the
            notification, which will be displayed below the title.
        :param badge: A USVString containing the URL of the image used to
            represent the notification when there is not enough space to
            display the notification itself.
        :param tag: A DOMString representing an identifying tag for the
            notification.
        :param icon: A USVString containing the URL of an icon to be
            displayed in the notification.
        :param image: a USVString containing the URL of an image to be
            displayed in the notification.
        :param vibrate: A vibration pattern for the device's vibration
            hardware to emit when the notification fires.
        :param require_interaction: Indicates that a notification should
            remain active until the user clicks or dismisses it, rather than
            closing automatically. The default value is false. (default=False)
        :param displayed: Set to true to display the notification.
        :param clicks: (default=0)
        :param clicks_timestamp: (default=-1)
        :param closes: Number of times the notification was closed.
            (default=0)
        :param closes_timestamp: (default=-1)
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Markdown(Component):
    """
    A react-markdown wrapper for dazzler.

    :CSS: ``dazzler-markdown-markdown``
    """
    _aspect_keys = []
    _children = []
    source = Aspect(required=True, docstring="Markdown string to render.")
    class_name = Aspect(docstring="CSS class of the container element.")
    style = Aspect(docstring="Style object to give to container.")
    escape_html = Aspect(default=True, docstring="Escape html element in the source. (default=True)")  # noqa: E501
    skip_html = Aspect(docstring="Skip rendering of html elements.")
    source_pos = Aspect(docstring="Add data-sourcepos attributes to elements.")
    allowed_types = Aspect(docstring="Which type of node to allow rendering, default to all.")  # noqa: E501
    disallowed_types = Aspect(docstring="Which types of nodes should not be rendered, default to none.")  # noqa: E501
    unwrap_disallowed = Aspect(docstring="Setting to true will try to extract/unwrap the childrenof disallowed nodes. For instance, if disallowing Strong,the default behaviour is to simply skip the text within thestrong altogether, while the behaviour some might want is tosimply have the text returned without the strong wrapping it.")  # noqa: E501
    link_target = Aspect(docstring="Set the default target for link tags. (`_blank` to open a new tab)")  # noqa: E501

    def __init__(
            self,
            source: str,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            escape_html: typing.Optional[bool] = True,
            skip_html: typing.Optional[bool] = UNDEFINED,
            source_pos: typing.Optional[bool] = UNDEFINED,
            allowed_types: typing.Optional[typing.List] = UNDEFINED,
            disallowed_types: typing.Optional[typing.List] = UNDEFINED,
            unwrap_disallowed: typing.Optional[bool] = UNDEFINED,
            link_target: typing.Optional[str] = UNDEFINED,
            identity: str = None
    ):
        """
        :param source: Markdown string to render.
        :param class_name: CSS class of the container element.
        :param style: Style object to give to container.
        :param escape_html: Escape html element in the source.
            (default=True)
        :param skip_html: Skip rendering of html elements.
        :param source_pos: Add data-sourcepos attributes to elements.
        :param allowed_types: Which type of node to allow rendering,
            default to all.
        :param disallowed_types: Which types of nodes should not be
            rendered, default to none.
        :param unwrap_disallowed: Setting to true will try to
            extract/unwrap the children of disallowed nodes. For instance, if
            disallowing Strong, the default behaviour is to simply skip the
            text within the strong altogether, while the behaviour some might
            want is to simply have the text returned without the strong
            wrapping it.
        :param link_target: Set the default target for link tags. (`_blank`
            to open a new tab)
        """
        Component.__init__(self, locals(), identity)



from dazzler.system import (
    Package as _Package,
    assets_to_requirements as _ass_to_req
)

from dazzler._assets import (
    assets as _assets,
    assets_dev as _dev,
    assets_dist_path as _dist_path,
    assets_dev_path as _dev_path,
)

from ._imports_ import *  # noqa: F401, F403
from ._imports_ import __all__

_name = 'core'
_package_name = f'dazzler_{_name}'


_components = []
for _c in __all__:
    _components.append(locals()[_c])

package = _Package(
    _package_name,
    components=_components,
    requirements=_ass_to_req(
        _dist_path, _assets.get(_name, {}),
        dev_data=_dev.get(_name, {}),
        dev_path=_dev_path,
        package_name=_package_name,
    )
)



import collections


class OrderedSet(collections.abc.MutableSet):
    """
    Slow set to keep the order of unique elements in a list.
    """
    def __init__(self, *args):
        self._data = []
        for i in args:
            self.add(i)

    # pylint: disable=arguments-differ
    def add(self, x):
        if x not in self._data:
            self._data.append(x)

    def discard(self, x):
        self._data.remove(x)

    def __contains__(self, x):
        return x in self._data

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        for i in self._data:
            yield i



import os
import json

from .tools import get_package_path

root_path = get_package_path('dazzler')
assets_path = os.path.abspath(os.path.join(root_path, 'assets'))
assets_dist_path = os.path.join(assets_path, 'dist')
assets_dev_path = os.path.join(assets_path, 'dev')
vendors_path = os.path.join(assets_path, 'vendors')
meta_path = os.path.join(assets_path, 'meta.js')

_asset_json = os.path.join(assets_dist_path, 'assets.json')
_asset_json_dev = os.path.join(assets_dev_path, 'assets.json')

if os.path.exists(_asset_json):
    with open(_asset_json) as _file:
        assets = json.load(_file)['chunks']
else:
    assets = {}

if os.path.exists(_asset_json_dev):
    with open(_asset_json_dev) as _file:
        assets_dev = json.load(_file)['chunks']
else:
    assets_dev = {}



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Grid(Component):
    """
    Render a list in a grid with a number of ``columns``.

    :CSS:

        ``dazzler-core-grid``
        - ``grid-row``
        - ``grid-cell``

    :Example:

    .. code-block:: python3

        from dazzler.components import core

        grid = core.Grid([1, 2, 3, 4], 2)
    """
    _aspect_keys = []
    _children = []
    children = Aspect(required=True, children=True, docstring="Children to render in a grid.")  # noqa: E501
    columns = Aspect(required=True, docstring="Number of columns")

    def __init__(
            self,
            children: typing.List[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]],  # noqa: E501
            columns: typing.Union[float, int],
            identity: str = None
    ):
        """
        :param children: Children to render in a grid.
        :param columns: Number of columns
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Slider(Component):
    """
    A slider with a caret contained within a min and max value.

    :CSS:

        ``dazzler-core-slider``
        - ``slider-area``: The outer div of the slider.
        - ``slider-handle``: The drag handle
        - ``slider-spacer``: The space that is before the handle.
    """
    _aspect_keys = []
    _children = []
    minimum = Aspect(required=True, docstring="Minimum (leftmost) value of the slider.")  # noqa: E501
    maximum = Aspect(required=True, docstring="Maximum (rightmost) value of the slider.")  # noqa: E501
    value = Aspect(default=0, docstring="Current value (default=0)")
    round = Aspect(docstring="Round the value (Possible values: 'ceil', 'floor')")  # noqa: E501
    debounce = Aspect(default=50, docstring="Time in milliseconds to wait before updating the value. (default=50)")  # noqa: E501
    class_name = Aspect(docstring="CSS classes to use. (Scope: dazzler-core-slider)")  # noqa: E501
    style = Aspect(docstring="Style object of the root div.")

    def __init__(
            self,
            minimum: typing.Union[float, int],
            maximum: typing.Union[float, int],
            value: typing.Optional[typing.Union[float, int]] = 0,
            round: typing.Optional[typing.Any] = UNDEFINED,
            debounce: typing.Optional[typing.Union[float, int]] = 50,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            identity: str = None
    ):
        """
        :param minimum: Minimum (leftmost) value of the slider.
        :param maximum: Maximum (rightmost) value of the slider.
        :param value: Current value (default=0)
        :param round: Round the value (Possible values: 'ceil', 'floor')
        :param debounce: Time in milliseconds to wait before updating the
            value. (default=50)
        :param class_name: CSS classes to use. (Scope: dazzler-core-slider)
        :param style: Style object of the root div.
        """
        Component.__init__(self, locals(), identity)



import secrets

from ._undefined import UNDEFINED


__all__ = [
    'Aspect', 'Component', 'prepare_aspects'
]


# noinspection PyProtectedMember
def prepare_aspects(aspects: dict):
    data = {}
    for aspect, value in aspects.items():
        if value is not UNDEFINED:
            if isinstance(value, Component):
                data[aspect] = value._prepare()
            elif isinstance(value, (list, tuple)):
                new_value = []
                for x in value:
                    if isinstance(x, Component):
                        new_value.append(x._prepare())
                    elif isinstance(x, dict):
                        new_value.append(prepare_aspects(x))
                    else:
                        new_value.append(x)
                data[aspect] = new_value
            elif isinstance(value, dict):
                data[aspect] = prepare_aspects(value)
            else:
                data[aspect] = value
    return data


def _generate_identity():
    return hex(secrets.randbits(40)).lstrip('0x')


# noinspection PyProtectedMember
class Aspect:
    """
    Aspects of component are represented on both backend and frontend.
    """
    def __init__(
            self,
            name=None,
            default=UNDEFINED,
            required=False,
            children=False,
            docstring=None,
    ):
        self.name = name
        self.default = default
        self.required = required
        self.children = children
        self.__doc__ = docstring

    def __set_name__(self, owner, name):
        self.name = name
        owner._aspects_keys.append(name)
        if self.children:
            owner._children.append(name)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        value = instance._aspects.get(self.name, UNDEFINED)
        return value

    def __set__(self, instance, value):
        if self.default is not UNDEFINED \
                and self.default == value \
                and not instance._initialized:
            return
        instance._aspects[self.name] = value


class Component:
    _aspects_keys = []
    _package_name = ''
    _children = []

    def __init__(
            self,
            aspects,
            identity: str = None,
    ):
        self.identity = identity or _generate_identity()
        self._initialized = False
        self._aspects = {}
        for k, v in aspects.items():
            if k in ('self', 'args', 'kwargs', 'identity'):
                continue
            setattr(self, k, v)
        self._initialized = True

    def _prepare(self) -> dict:
        return {
            'name': self.__class__.__name__,
            'identity': self.identity,
            'package': self._package_name,
            'aspects': prepare_aspects(self._aspects),
        }

    def _paths(self, parent=''):
        path = f'{parent + "." if parent else ""}{self.identity}'
        yield path, self
        for key in self._children:
            value = getattr(self, key, UNDEFINED)
            if value is not UNDEFINED:
                if isinstance(value, Component):
                    for p, nested in value._paths(path):
                        yield p, nested
                elif isinstance(value, list):
                    for component in (
                            x for x in value if isinstance(x, Component)
                    ):
                        # noinspection PyProtectedMember
                        for p, nested in component._paths(path):
                            yield p, nested
                else:
                    yield f'{path}.{key}', value

    def __repr__(self):
        return f'<[{self._package_name}.{self.__class__.__name__}]' \
            f' ({self.identity})>'



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Spinner(Component):
    """
    Simple html/css spinner.
    """
    _aspect_keys = []
    _children = []
    class_name = Aspect()
    style = Aspect()

    def __init__(
            self,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            identity: str = None
    ):
        """
        :param class_name:
        :param style:
        """
        Component.__init__(self, locals(), identity)



import typing


from dazzler.errors import PackageConflictError

from ._requirements import Requirement
from ._component import Component


class Package:
    package_registry = {}

    def __init__(
            self,
            name: str,
            components: typing.List[Component] = None,
            requirements: typing.List[Requirement] = None,
            routes: list = None,
    ):
        if name in self.package_registry:
            raise PackageConflictError(f'Duplicate package name: {name}')

        self.name = name
        self.components = {}

        for component in (components or []):
            setattr(component, '_package_name', self.name)

            self.components[component.__name__] = component

        self.requirements = requirements or []
        self.routes = routes or []
        self.package_registry[self.name] = self

    def prepare(self, dev=False, external=False):
        return {
            'name': self.name,
            'components': list(self.components.keys()),
            'requirements': [
                x.prepare(dev=dev, external=external)
                for x in self.requirements
            ],
        }

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'<Package "{self.name}">'



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Dropdown(Component):
    """
    A dropdown to select options from a list.

    :Example:

    .. code-block:: python3

         from dazzler.components.core import Dropdown

         dropdown = Dropdown(
             {'label': 'Option 1', 'value': 1},
                {'label': 'Option 2', 'value': 2}
         )
    """
    _aspect_keys = []
    _children = []
    options = Aspect(required=True, children=True, docstring="List of options to choose/search from.")  # noqa: E501
    value = Aspect(docstring="Currently selected value(s).")
    multi = Aspect(docstring="Allow multiple values to be chosen, the value become a list of values.")  # noqa: E501
    searchable = Aspect(docstring="If true, render an input")
    search_value = Aspect(default='', docstring="Value entered by user to search options. (default='')")  # noqa: E501
    search_props = Aspect(docstring="Keys to filter on searching the options.- Leave empty for all props.- Valid values types to search on are strings & arrays.- Nested prop access with dot notation.")  # noqa: E501
    search_label = Aspect(docstring="Search the label along with the value.")
    search_backend = Aspect(docstring="Do not perform any search on the options from the frontend and insteadrelies on binding the ``search_value`` to filter and set the``filtered_options`` aspect.")  # noqa: E501
    filtered_options = Aspect(default=UNDEFINED, docstring="Array of options that are filtered, set from backend with search (default=null)")  # noqa: E501
    opened = Aspect(default=False, docstring="Is the dropdown currently open ? (default=False)")  # noqa: E501
    no_results_label = Aspect(default=UNDEFINED, children=True, docstring="Label to use when no search results are available.")  # noqa: E501
    toggle_symbol = Aspect(default='⏷', docstring="Unicode character used as the toggle button. (default='⏷')")  # noqa: E501
    scrollable = Aspect(docstring="Make the menu scrollable.")
    scroll_max_size = Aspect()
    class_name = Aspect()
    style = Aspect()

    def __init__(
            self,
            options: typing.List[typing.Union[typing.Dict[str, typing.Union[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]], typing.Union[str, typing.Union[float, int], typing.Dict], typing.Dict, str]], str]],  # noqa: E501
            value: typing.Optional[typing.Union[str, typing.List, typing.Union[float, int], typing.Dict]] = UNDEFINED,  # noqa: E501
            multi: typing.Optional[bool] = UNDEFINED,
            searchable: typing.Optional[bool] = UNDEFINED,
            search_value: typing.Optional[str] = '',
            search_props: typing.Optional[typing.List[str]] = UNDEFINED,
            search_label: typing.Optional[bool] = UNDEFINED,
            search_backend: typing.Optional[bool] = UNDEFINED,
            filtered_options: typing.Optional[typing.List] = UNDEFINED,
            opened: typing.Optional[bool] = False,
            no_results_label: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            toggle_symbol: typing.Optional[str] = '⏷',
            scrollable: typing.Optional[bool] = UNDEFINED,
            scroll_max_size: typing.Optional[typing.Union[float, int]] = UNDEFINED,  # noqa: E501
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            identity: str = None
    ):
        """
        :param options: List of options to choose/search from.
        :param value: Currently selected value(s).
        :param multi: Allow multiple values to be chosen, the value become
            a list of values.
        :param searchable: If true, render an input
        :param search_value: Value entered by user to search options.
            (default='')
        :param search_props: Keys to filter on searching the options.  -
            Leave empty for all props. - Valid values types to search on are
            strings & arrays. - Nested prop access with dot notation.
        :param search_label: Search the label along with the value.
        :param search_backend: Do not perform any search on the options
            from the frontend and instead relies on binding the
            ``search_value`` to filter and set the ``filtered_options`` aspect.
        :param filtered_options: Array of options that are filtered, set
            from backend with search (default=null)
        :param opened: Is the dropdown currently open ? (default=False)
        :param no_results_label: Label to use when no search results are
            available.
        :param toggle_symbol: Unicode character used as the toggle button.
            (default='⏷')
        :param scrollable: Make the menu scrollable.
        :param scroll_max_size:
        :param class_name:
        :param style:
        """
        Component.__init__(self, locals(), identity)



import functools
import importlib
import itertools
import shutil
import sys
import asyncio
import os
import typing

from concurrent.futures import ThreadPoolExecutor

import appdirs
import precept

from .system.auth import Authenticator, DazzlerAuth, AuthBackend
from .tools import get_member
from .system.session import (
    SessionMiddleware, FileSessionBackEnd, RedisSessionBackend
)
from .system import (
    Package,
    generate_components,
    generate_meta,
    Requirement,
    Page,
    Middleware,
    Route,
    RouteMethod,
)

from ._config import DazzlerConfig
from ._server import Server
from ._version import __version__
from .errors import (
    PageConflictError, ServerStartedError, SessionError, AuthError,
    NoInstanceFoundError
)
from ._assets import assets_path
from ._reloader import start_reloader


class Dazzler(precept.Precept):  # pylint: disable=too-many-instance-attributes
    """
    Dazzler web application & cli.

    :Commands:

        - ``dazzler --root application``, start an application.
        - ``dazzler generate components_metadata.json output_dir``
        - ``dazzler copy-requirements -p dazzler_core -p dazzler_extra``
    """
    config: DazzlerConfig
    config_class = DazzlerConfig
    version = __version__
    pages: typing.Dict[str, Page]
    global_arguments = [
        precept.Argument(
            '-a', '--application',
            help='Path to the application',
            type=str
        ),
        precept.Argument(
            '--reloaded',
            action='store_true'
        ),
    ]
    server: Server
    auth: DazzlerAuth

    def __init__(self, module_name, app_name=None):
        module = importlib.import_module(module_name)
        self.module_file = module.__file__
        self.module_name = module_name
        self.root_path = os.path.dirname(module.__file__)
        self.app_name = app_name or \
            os.path.basename(module.__file__).rstrip('.py')

        super().__init__(
            config_file=[
                'dazzler.toml',
                os.path.join(self.root_path, 'dazzler.toml')
            ],
            add_dump_config_command=True,
            executor=ThreadPoolExecutor(),
            print_version=False,
        )

        self.requirements: typing.List[Requirement] = []
        self.middlewares: typing.List[Middleware] = []
        self.server = Server(self, loop=self.loop)
        self.pages = {}
        self.stop_event = asyncio.Event()
        self._prepared = False
        self._started = False
        self.data_dir = os.path.join(
            appdirs.user_data_dir('dazzler'), self.app_name
        )
        self.requirements_dir = os.path.join(self.data_dir, 'requirements')

    def add_page(self, *pages: Page):
        if self._started:
            raise ServerStartedError(
                f'Cannot add pages after the server started:'
                f' {",".join(str(x) for x in pages)}'
            )
        for page in pages:
            if page.name in self.pages:
                raise PageConflictError(f'Duplicate page name: {page}')
            for x in self.pages.values():
                if x.url == page.url:
                    raise PageConflictError(f'Duplicate page url: {page}')
            self.pages[page.name] = page

    async def stop(self):
        if self.server.site:
            await self.server.site.stop()
        self.stop_event.set()

    # pylint: disable=arguments-differ
    async def main(
            self,
            application=None,
            blocking=True,
            debug=False,
            reload=False,
            reloaded=False,
            start_event=None,
            **kwargs
    ):
        if application:
            # When running from the command line need to insert the path.
            # Otherwise it will never find it.
            sys.path.insert(0, '.')
            module = importlib.import_module(application)

            app: typing.Optional[Dazzler] = None

            # Search for instance so you don't have to supply a name
            # with weird syntax.
            for instance in vars(module).values():
                if isinstance(instance, Dazzler):
                    app: Dazzler = instance

            if app is None:
                raise NoInstanceFoundError(
                    f'No dazzler application found in {application}'
                )

            # Set cli for arguments
            app.cli = self.cli
            await app._on_parse(self._args)
        else:
            app = self

        reloader = None

        if reload:
            app.config.development.reload = True
            reloader = self.loop.create_task(
                start_reloader(app, reloaded, start_event)
            )

        if not reload or (reload and reloaded):
            if debug:
                app.config.debug = debug
            await app.setup_server(debug=app.config.debug)

            await app.server.start(
                app.config.host, app.config.port,
            )

        # Test needs it to run without loop
        # Otherwise the server closes when the event loop ends.
        if blocking and not reloader:
            await self._loop()
        elif blocking and reloader:
            await reloader

    async def _loop(self):
        while not self.stop_event.is_set():
            # Just run so the site doesn't close.
            # Maybe add something later here.
            await asyncio.sleep(100)

    async def setup_server(self, debug=False):
        await self._handle_configs()

        # Copy all requirements to make sure all is latest.
        await self.copy_requirements()

        await self.events.dispatch('dazzler_setup', application=self)

        # Add package defined routes
        routes = [x.routes for x in Package.package_registry.values()]

        # Add page specifics routes
        for page in self.pages.values():
            # Pages are not prefixed
            # noinspection PyTypeChecker
            routes += [[
                Route(
                    page.url,
                    functools.partial(self.server.route_page, page=page),
                    name=page.name,
                ),
                Route(
                    page.url,
                    functools.partial(self.server.route_page_json, page=page),
                    name=f'{page.name}-api',
                    method=RouteMethod.POST,
                ),
                Route(
                    f'/{page.name}/ws',
                    functools.partial(self.server.route_update, page=page),
                    name=f'{page.name}-ws',
                    method=RouteMethod.GET,
                ),
            ]]
            routes += [page.routes]

        self.server.setup_routes(
            list(itertools.chain(*routes)),
            debug=debug,
        )

        self._prepared = True
        self.server.app.on_startup.append(self._on_startup)
        self.server.app.on_shutdown.append(self._on_shutdown)

    async def application(self):
        """Call for wsgi application"""
        if not self._prepared:
            await self.setup_server()
        return self.server.app

    def collect_requirements(self) -> typing.Set[Requirement]:
        requirements = [
            package.requirements for package
            in Package.package_registry.values()
        ]
        requirements += [self.requirements]
        requirements += [
            page.requirements for page in self.pages.values()
        ]

        return set(itertools.chain(*requirements))

    def requirement_from_file(self, file) -> typing.Optional[Requirement]:
        for requirement in self.collect_requirements():
            if any(
                    file in r for r in (requirement.internal, requirement.dev)
                    if r
            ):
                return requirement
        return None

    def _remove_requirement(self, requirement: Requirement):
        # Remove from page requirement most probably
        for page in self.pages.values():
            if requirement in page.requirements:
                page.requirements.remove(requirement)
                return
        if requirement in self.requirements:
            self.requirements.remove(requirement)
            return
        for package in Package.package_registry.values():
            if requirement in package.requirements:
                package.requirements.remove(requirement)
                return

    # Commands

    @precept.Command(
        precept.Argument(
            'source_directory',
            help='Components source directory'
        ),
        precept.Argument(
            'output_dir',
            help='Output the components in this folder.'
        ),
        description='Generate dazzler components from react-docgen output'
    )
    async def generate(self, source_directory, output_dir):
        os.makedirs(output_dir, exist_ok=True)
        metadata = await generate_meta(source_directory)
        await generate_components(metadata, output_dir, self.executor)

    @precept.Command(
        precept.Argument('-p', '--packages', nargs='+'),
        description='Copy the application requirements.'
    )
    async def copy_requirements(self, packages=tuple()):
        self.logger.debug('Copying requirements.')

        if not os.path.exists(self.data_dir):
            os.makedirs(self.data_dir)

        shutil.rmtree(self.requirements_dir, ignore_errors=True)
        os.makedirs(self.requirements_dir, exist_ok=True)

        for package in packages:
            self.logger.debug(f'Importing package: {package}')
            importlib.import_module(package)

        futures = []

        for requirement in self.collect_requirements():
            if not requirement.internal:
                continue
            destination = os.path.join(
                self.requirements_dir, requirement.internal_static
            )
            self.logger.debug(
                f'Copying {requirement.internal} '
                f'to {destination}'
            )
            os.makedirs(os.path.dirname(destination), exist_ok=True)
            futures.append(self.executor.execute(
                shutil.copy,
                requirement.internal,
                destination,
            ))
            if requirement.dev:
                destination = os.path.join(
                    self.requirements_dir, requirement.dev_static
                )
                self.logger.debug(
                    f'Copying {requirement.dev} '
                    f'to {destination}'
                )
                os.makedirs(os.path.dirname(destination), exist_ok=True)
                futures.append(
                    self.executor.execute(
                        shutil.copy,
                        requirement.dev,
                        destination,
                    )
                )

        # Copy the mount script.
        futures.append(
            self.executor.execute(
                shutil.copy,
                os.path.join(assets_path, 'index.js'),
                self.requirements_dir
            )
        )
        await asyncio.gather(*futures)

    # Handlers

    async def on_file_change(
            self,
            filenames: typing.List[str],
            deleted: typing.List[str]
    ):
        hot = False  # Restart the server & Reload the page api/layout.
        refresh = False  # Refresh the page because the root bundles changed.
        files = set()
        deleted_files = set()
        for filename in filenames:
            if filename.endswith('.py'):
                hot = True
            elif filename.endswith('.js') or filename.endswith('.css'):
                if 'dazzler_renderer' in filename:
                    refresh = True
                requirement = self.requirement_from_file(filename)
                if requirement:
                    files.add(requirement)
                else:
                    requirement = Requirement(internal=filename)
                    files.add(requirement)
                    self.requirements.append(requirement)

        for removed in deleted:
            requirement = self.requirement_from_file(removed)
            if removed.endswith('.js'):
                refresh = True
            if requirement:
                self._remove_requirement(requirement)
                deleted_files.add(requirement)

        if not hot:
            await self.copy_requirements()
        await self.server.send_reload(
            [r.prepare(dev=self.config.debug) for r in files],
            hot,
            refresh,
            [r.prepare(dev=self.config.debug) for r in deleted_files]
        )
        return hot

    async def _on_startup(self, _):
        await self.events.dispatch('dazzler_start', application=self)

    async def _on_shutdown(self, _):
        self.stop_event.set()
        await self.events.dispatch('dazzler_stop', application=self)

    async def _enable_auth(self):
        self.logger.debug('Enabling authentication system')
        backend = None

        if self.config.authentication.authenticator:
            authenticator = get_member(
                self.config.authentication.authenticator
            )
            if isinstance(authenticator, type):
                authenticator = authenticator()

            if not isinstance(authenticator, Authenticator):
                raise AuthError(
                    f'{self.config.authentication.authenticator} '
                    f'is not an instance of '
                    f'`dazzler.system.auth.Authenticator`'
                    f'{repr(authenticator)}'
                )
        else:
            raise AuthError(
                'No authenticator provided in config'
            )

        if self.config.authentication.backend:
            backend = get_member(
                self.config.authentication.backend,
            )
            if isinstance(backend, type):
                backend = backend()

            if not isinstance(backend, AuthBackend):
                raise AuthError(
                    f'{self.config.authentication.backend} '
                    f'is not an instance of '
                    f'`dazzler.system.auth.AuthBackend`'
                )

        self.auth = DazzlerAuth(self, authenticator, backend=backend)

    async def _handle_configs(self):
        # Gather global requirements from configs.
        for external in itertools.chain(
                self.config.requirements.external_scripts,
                self.config.requirements.external_styles
        ):
            self.requirements.append(Requirement(external=external))

        for internal in itertools.chain(
                self.config.requirements.internal_scripts,
                self.config.requirements.internal_styles
        ):
            self.requirements.append(Requirement(internal=internal))

        if self.config.session.enable:

            if self.config.session.backend == 'File':
                backend = FileSessionBackEnd(self)
            elif self.config.session.backend == 'Redis':
                backend = RedisSessionBackend(self)
            else:
                raise SessionError(
                    'No valid session backend defined.\n',
                    'Please choose from "File" or "Redis"'
                )

            self.middlewares.insert(
                0, SessionMiddleware(self, backend=backend)
            )

        if self.config.authentication.enable:
            await self._enable_auth()


def cli():
    dazzler = Dazzler('__main__')
    dazzler.start()



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Button(Component):
    """
    A button to click on!

    :CSS: ``dazzler-core-button``

    :Example:

    .. code-block:: python3

        from dazzler.system import Page, Trigger
        from dazzler.components import core

        button = core.Button('Click me', identity='btn')

        page = Page(
            __name__,
            core.Container([button, core.Container(identity="output")])
        )

        @page.binding(Trigger('btn', 'clicks'))
        async def on_click(ctx):
            await ctx.set_aspect('output', f'Clicked {ctx.trigger.value}')
    """
    _aspect_keys = []
    _children = []
    label = Aspect(required=True, children=True, docstring="Text or component to display.")  # noqa: E501
    clicks = Aspect(default=0, docstring="The number of times the button was clicked. (default=0)")  # noqa: E501
    class_name = Aspect(docstring="Class name to give the button.")
    style = Aspect(docstring="Style object to give to the Button.")
    id = Aspect(docstring="DOM id, otherwise the identity is used.")
    preset = Aspect(docstring="Preset style colors to apply. (Possible values: 'primary', 'secondary', 'danger', 'warning', 'success')")  # noqa: E501
    disabled = Aspect(docstring="Disable the button.")
    rounded = Aspect(docstring="Round the edges.")
    circle = Aspect(docstring="Circle button")
    bordered = Aspect(default=True, docstring="Add a border (default=True)")
    size = Aspect(docstring="The size of the button. (Possible values: 'tiny', 'small', 'medium', 'large', 'larger', 'x-large', 'xx-large')")  # noqa: E501

    def __init__(
            self,
            label: typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]],  # noqa: E501
            clicks: typing.Optional[typing.Union[float, int]] = 0,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            id: typing.Optional[str] = UNDEFINED,
            preset: typing.Optional[typing.Any] = UNDEFINED,
            disabled: typing.Optional[bool] = UNDEFINED,
            rounded: typing.Optional[bool] = UNDEFINED,
            circle: typing.Optional[bool] = UNDEFINED,
            bordered: typing.Optional[bool] = True,
            size: typing.Optional[typing.Any] = UNDEFINED,
            identity: str = None
    ):
        """
        :param label: Text or component to display.
        :param clicks: The number of times the button was clicked.
            (default=0)
        :param class_name: Class name to give the button.
        :param style: Style object to give to the Button.
        :param id: DOM id, otherwise the identity is used.
        :param preset: Preset style colors to apply. (Possible values:
            'primary', 'secondary', 'danger', 'warning', 'success')
        :param disabled: Disable the button.
        :param rounded: Round the edges.
        :param circle: Circle button
        :param bordered: Add a border (default=True)
        :param size: The size of the button. (Possible values: 'tiny',
            'small', 'medium', 'large', 'larger', 'x-large', 'xx-large')
        """
        Component.__init__(self, locals(), identity)



from enum import auto
from typing import Callable, Awaitable, Union

from aiohttp import web
from precept import AutoNameEnum


class RouteMethod(AutoNameEnum):
    GET = auto()
    POST = auto()
    PUT = auto()
    PATCH = auto()
    DELETE = auto()
    CONNECT = auto()
    OPTIONS = auto()
    HEAD = auto()

    def get_method(self):
        return getattr(web, self.value)


class Route:
    def __init__(
            self,
            path: str,
            handler: Callable[[web.Request], Awaitable[web.Response]],
            method: Union[str, RouteMethod] = RouteMethod.GET,
            name: str = None
    ):
        self.path = path
        self.handler = handler
        if isinstance(method, str):
            self.method = getattr(RouteMethod, method.upper())
        else:
            self.method = method
        self.name = name



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class PopUp(Component):
    """
    Wraps a component/text to render a popup when hovering
    over the children or clicking on it.
    """
    _aspect_keys = []
    _children = []
    children = Aspect(children=True, docstring="Component/text to wrap with a popup on hover/click.")  # noqa: E501
    content = Aspect(children=True, docstring="Content of the popup info.")
    active = Aspect(default=False, docstring="Is the popup currently active. (default=False)")  # noqa: E501
    mode = Aspect(default="'hover'", docstring="Show popup on hover or click. (Possible values: 'hover', 'click') (default="'hover'")")  # noqa: E501
    class_name = Aspect(docstring="CSS for the wrapper.")
    style = Aspect(docstring="Style of the wrapper.")
    content_style = Aspect(docstring="Style for the popup.")
    children_style = Aspect(docstring="Style for the wrapped children.")

    def __init__(
            self,
            children: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            content: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            active: typing.Optional[bool] = False,
            mode: typing.Optional[typing.Any] = "'hover'",
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            content_style: typing.Optional[typing.Dict] = UNDEFINED,
            children_style: typing.Optional[typing.Dict] = UNDEFINED,
            identity: str = None
    ):
        """
        :param children: Component/text to wrap with a popup on
            hover/click.
        :param content: Content of the popup info.
        :param active: Is the popup currently active. (default=False)
        :param mode: Show popup on hover or click. (Possible values:
            'hover', 'click') (default="'hover'")
        :param class_name: CSS for the wrapper.
        :param style: Style of the wrapper.
        :param content_style: Style for the popup.
        :param children_style: Style for the wrapped children.
        """
        Component.__init__(self, locals(), identity)



import os
import typing
from aiohttp import web

from ..tools import get_package_path
from ._binding import Binding, BoundAspect, Trigger, State
from ._component import Component
from ._package import Package
from ._requirements import Requirement, collect_requirements
from ._route import Route, RouteMethod
from ..errors import BindingError


def _coerce_binding(value, binding_type: typing.Type = Trigger):

    if isinstance(value, list):
        return [_coerce_binding(x, binding_type) for x in value]

    if isinstance(value, binding_type):
        return value

    if isinstance(value, str):

        splat = value.split('@')
        if not len(splat) == 2:
            raise BindingError(f'Invalid {binding_type.__name__}: {value}')

        return binding_type(splat[1], splat[0])

    raise BindingError(f'Invalid {binding_type.__name__}: {value}')


# pylint: disable=too-many-instance-attributes
class Page:
    """A page of the application."""

    def __init__(
            self,
            name: str,
            layout: typing.Union[Component, typing.Callable[[web.Request], typing.Awaitable[Component]]],  # noqa: E501
            url: str = None,
            bindings: list = None,
            routes: typing.List[Route] = None,
            requirements: typing.List[Requirement] = None,
            requirements_dir: str = 'requirements',
            title: str = None,
            lang: str = 'en',
            header: str = '',
            footer: str = '',
            favicon: str = '',
            meta_tags: typing.List[typing.Dict[str, str]] = None,
            packages: typing.List[str] = None,
            require_login: bool = False,
    ):
        """
        :param name: Unique name for the page, usually give __name__.
        :param layout: Root component or function returning a Component. In the
            case of a function, the request will be None on setup.
        :param url: Url, default to ``/name``.
        :param bindings: Bindings of component aspects for this page.
        :param routes: Additional routes to register with this page.
        :param requirements: List of requirements to
        :param title: Title of the page.
        :param lang: lang attribute of the html tag.
        :param header: Static header to include as first child on the page.
        :param footer: Static footer to include as last child on the page.
        :param favicon: Url to favicon for the page.
        :param meta_tags: Meta tags of the page.
        :param packages: Packages list to use on the page instead
            of the whole registry.
        :param require_login: Page requires that user is logged in.
        """
        self.name = name.split('.')[-1]
        self.base_name = name
        try:
            self.page_path = get_package_path(name)
        except ImportError:
            # The requirements dir is then absolute.
            self.page_path = ''

        self.title = title or self.name
        self.url = url or f'/{self.name}'
        self.routes = routes or []
        self.layout = layout
        self.bindings = bindings or []
        self.requirements = requirements or []
        for requirement in self.requirements:
            requirement.page = self.name

        self.requirements_dir = os.path.join(self.page_path, requirements_dir)

        # Collect all requirement in the requirement folder.
        if os.path.exists(self.requirements_dir):
            self.requirements += collect_requirements(
                self.requirements_dir, page=self.name
            )

        self.header = header
        self.footer = footer
        self.favicon = favicon
        self.meta_tags = meta_tags or []
        self.packages = packages
        self.lang = lang
        self._bindings = {str(x.trigger): x for x in self.bindings}
        self.require_login = require_login

    async def prepare(
            self,
            request: web.Request,
            debug=False,
            external=False
    ) -> dict:
        """
        Prepare the page for rendering.

        :param request: The request to prepare for.
        :param debug: To collect dev package
        :param external: Serve external requirements if available.
        :return: prepared dict with layout and page name.
        """
        layout = self.layout
        if callable(layout):
            layout = await layout(request)

        packages = [
            Package.package_registry[x].prepare(dev=debug)
            for x in self.packages
        ] if self.packages is not None else [
            x.prepare(dev=debug, external=external)
            for x in Package.package_registry.values()
            if x.name != 'dazzler_renderer'
        ]

        # noinspection PyProtectedMember
        return {
            'layout': layout._prepare(),
            'page': self.name,
            'bindings': {
                x['key']: x for y in self.bindings for x in y.prepare()
            },
            'requirements': [
                x.prepare(dev=debug, external=external)
                for x in self.requirements
            ],
            'packages': packages
        }

    def bind(
            self,
            trigger: typing.Union[Trigger, str],
            *states: typing.Union[State, str]
    ):
        """
        Attach a function to be called when the trigger update.

        :param trigger:
        :param states:
        :return:
        """
        trg = _coerce_binding(trigger)
        sts = _coerce_binding(list(states), State)

        def _wrapper(func):
            binding = Binding(trg, sts)(func)
            self.bindings.append(binding)
            for trig in binding.triggers:
                self._bindings[str(trig)] = binding
            return func

        return _wrapper

    def get_binding(self, key) -> BoundAspect:
        """
        Get the binding from the trigger key.

        :param key: Trigger key
        :return:
        """
        return self._bindings.get(key)

    def route(
            self,
            path,
            method: typing.Union[str, RouteMethod] = RouteMethod.GET,
            name=None,
            prefix=True
    ):
        """
        Add a route with a decorator.

        :param path: Url to handle.
        :param method: Method of the route.
        :param name: Unique name for the route.
        :param prefix: Prefix the path with the page path.
        :return:
        """
        if prefix:
            url = '/'.join([self.url.rstrip('/'), path.lstrip('/')])
        else:
            url = path

        def _page_route(func):
            self.routes.append(
                Route(url, func, name=name, method=method)
            )
            return func

        return _page_route

    def __str__(self):
        return f'{self.name}@{self.url}'

    def __hash__(self):
        return hash((self.name, self.url))

    def __repr__(self):
        return f'<Page name="{self.name}" href="{self.url}">'



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Sticky(Component):
    """
    A shorthand component for a sticky div.
    """
    _aspect_keys = []
    _children = []
    children = Aspect(children=True)
    top = Aspect()
    left = Aspect()
    right = Aspect()
    bottom = Aspect()
    class_name = Aspect()
    style = Aspect()

    def __init__(
            self,
            children: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            top: typing.Optional[str] = UNDEFINED,
            left: typing.Optional[str] = UNDEFINED,
            right: typing.Optional[str] = UNDEFINED,
            bottom: typing.Optional[str] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            identity: str = None
    ):
        """
        :param children:
        :param top:
        :param left:
        :param right:
        :param bottom:
        :param class_name:
        :param style:
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class ViewPort(Component):
    """
    Activate a view with key.

    Standalone or tabbed.

    :CSS:

        ``dazzler-core-view-port``
        - ``view-content``
        - ``dazzler-tab``
        - ``dazzler-tabs``
        - ``tab-active``
    """
    _aspect_keys = []
    _children = []
    active = Aspect(required=True, docstring="The view that is active.")
    views = Aspect(required=True, children=True, docstring="A dictionary of components to render with the active key.")  # noqa: E501
    class_name = Aspect(docstring="The class name of the outer div.")
    style = Aspect(docstring="Style object of the outer div.")
    tabbed = Aspect(docstring="Incorporate tabs before the views.")
    tab_labels = Aspect(docstring="Labels for the tabs otherwise the tab keys will be used.")  # noqa: E501
    vertical_tabs = Aspect(docstring="Make the tabs vertical aligned.")
    rounded_tabs = Aspect(docstring="Round tab style")
    bordered = Aspect(docstring="Add a border around the viewport content (CSS bordered).")  # noqa: E501

    def __init__(
            self,
            active: str,
            views: typing.Dict[str, typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]],  # noqa: E501
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            tabbed: typing.Optional[bool] = UNDEFINED,
            tab_labels: typing.Optional[typing.Dict[str, str]] = UNDEFINED,
            vertical_tabs: typing.Optional[bool] = UNDEFINED,
            rounded_tabs: typing.Optional[bool] = UNDEFINED,
            bordered: typing.Optional[bool] = UNDEFINED,
            identity: str = None
    ):
        """
        :param active: The view that is active.
        :param views: A dictionary of components to render with the active
            key.
        :param class_name: The class name of the outer div.
        :param style: Style object of the outer div.
        :param tabbed: Incorporate tabs before the views.
        :param tab_labels: Labels for the tabs otherwise the tab keys will
            be used.
        :param vertical_tabs: Make the tabs vertical aligned.
        :param rounded_tabs: Round tab style
        :param bordered: Add a border around the viewport content (CSS
            bordered).
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Store(Component):
    """
    Store data in the browser memory. Data last until the page is refreshed.
    """
    _aspect_keys = []
    _children = []
    data = Aspect(docstring="Stored data.")

    def __init__(
            self,
            data: typing.Optional[typing.Any] = UNDEFINED,
            identity: str = None
    ):
        """
        :param data: Stored data.
        """
        Component.__init__(self, locals(), identity)



import itertools
import os
import asyncio
import sys
import typing
import pathlib
import subprocess
import time
import re

from ._assets import assets_path
from .tools import OrderedSet


# aioredis likes to changes it's files at runtime...
redis_exclude = re.compile(r'(stringprep\.py|aioredis|hiredis|idna\.py)')


async def watch(
        directories: typing.Set,
        extra_files: list,
        on_change: typing.Callable,
        interval: float = 0.5,
        threshold: float = 3.0,
):
    """
    Watch for file changes in directories and files.

    :param directories: Directories to watch.
    :param extra_files:
    :param on_change: What to do on change, takes the names of the changed
        files as first argument.
    :param interval: The rate at which to perform the change check on the files
    :param threshold: Wait time from first change til it fires reload events.
    :return:
    """
    timestamps = {}
    initial = True

    def handle_file(filepath):
        path = pathlib.Path(filepath)

        # Removes eggs
        if '.egg/' in filepath:
            return False

        if not path.exists():
            if filepath in timestamps:
                timestamps.pop(filepath)
                return True
            return False

        new = path.stat().st_mtime
        old = timestamps.get(filepath)

        if old != new:
            timestamps[filepath] = new
            if not initial:
                return True
        return False

    change_start = None
    changed = set()
    deleted_files = set()

    while True:
        looped = set()
        for file in itertools.chain(
                (
                    os.path.join(current, f)
                    for directory in directories
                    for current, _, files in os.walk(directory)
                    for f in files
                ),
                # Python files, won't detect new files unless they are imported
                # by a file that is currently used.
                (
                    x.__file__
                    for x in sys.modules.values()
                    if (
                            hasattr(x, '__file__')
                            and x.__file__
                            and not redis_exclude.search(x.__file__)
                    )
                ),
                extra_files,
        ):
            looped.add(file)
            if handle_file(file):
                changed.add(file)

        missings = set(timestamps.keys()).difference(looped)
        for removed in missings:
            deleted_files.add(removed)
            timestamps.pop(removed)

        if changed or deleted_files:
            if change_start is None:
                change_start = time.time()

            if time.time() - change_start > threshold:
                stop = await on_change(changed, deleted_files)
                if stop:
                    return
                changed = set()
                deleted_files = set()
                change_start = None

        initial = False
        await asyncio.sleep(interval)


async def run_reloaded(app, start_event):
    run_stop = asyncio.Event()

    if app.module_name == '__main__':
        # Main can't be called so get the path from the file instead.
        app_path = pathlib.Path(app.module_file).relative_to(os.getcwd())
        app_path = str(app_path).replace('.py', '').replace(os.sep, '.')
    else:
        app_path = app.module_name

    app.logger.debug(f'Starting app in reloaded mode: {app_path}')

    args = OrderedSet(*f'{sys.executable} -m dazzler'.split())
    for arg in itertools.chain(
            app.cli.raw_args,
            ['--application', app_path, '--reload', '--reloaded']
    ):
        args.add(arg)

    proc = subprocess.Popen(
        list(args),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    def handle_std(stream, outfile):
        while not run_stop.is_set() and not app.stop_event.is_set():
            data = stream.readline()
            if app.stop_event.is_set():
                proc.kill()
                proc.wait()
                return
            if data:
                data = data.decode().strip()
                if start_event and 'Started server' in data:
                    start_event.set()
                print(data, file=outfile)

    async def handle_poll():
        while proc.poll() is None:
            if app.stop_event.is_set():
                proc.kill()
                proc.wait()
                run_stop.set()
                return
            await asyncio.sleep(0.001)
        run_stop.set()

    async def cleanup(_):
        if proc.poll() is None:
            proc.kill()
            proc.wait()

    async def handle_stop():
        await app.stop_event.wait()
        await cleanup(None)

    # Keyboard interrupt prevent the subprocess from terminating and
    # would otherwise require a KILL signal.
    app.events.subscribe('KeyboardInterrupt', cleanup)
    app.events.subscribe('dazzler_stop', cleanup)
    stopper = app.loop.create_task(handle_stop())

    await asyncio.gather(
        # Run these in threads since they block and get stuck
        app.executor.execute(handle_std, proc.stdout, None),
        app.executor.execute(handle_std, proc.stderr, sys.stderr),
        handle_poll(),
    )
    if not stopper.done():
        stopper.cancel()


async def start_reloader(app, reloaded=False, start_event=None):
    """
    Start the reload mechanism.

    Re-run the application under a new subprocess if root.
    Otherwise run the watch and send reload events to connected
    websockets.

    :param app: The dazzler instance to reload
    :type app: dazzler.Dazzler
    :param reloaded: If the python is currently reloaded.
    :param start_event: Event to set once the reloaded instance has started.
    :return:
    """

    if not reloaded:
        while not app.stop_event.is_set():
            # Run this forever reloading every time the watch detect python
            # file change
            await run_reloaded(app, start_event)
            app.logger.info('Reloading...')
    else:
        # Watch files on live server to send reload update thru ws.

        watch_directories = set(
            directory for directory in itertools.chain(
                [app.config.static_folder, assets_path],
                [x.requirements_dir for x in app.pages.values()]
            ) if directory
        )
        # Take all requirements that are not in the watch directories.
        extra_files = [
            requirement.internal
            for requirement in app.collect_requirements()
            if requirement.internal and not any(
                pathlib.Path(x) in pathlib.Path(requirement.internal).parents
                for x in watch_directories
            )
        ]
        app.logger.debug('Starting reload watch.')
        await watch(
            watch_directories,
            extra_files,
            app.on_file_change,
            app.config.development.reload_interval,
            app.config.development.reload_threshold,
        )
        await app.stop()



from ._drawer import Drawer
from ._notice import Notice
from ._pager import Pager
from ._pop_up import PopUp
from ._spinner import Spinner
from ._sticky import Sticky
from ._tree_view import TreeView

__all__ = [
    "Drawer",
    "Notice",
    "Pager",
    "PopUp",
    "Spinner",
    "Sticky",
    "TreeView"
]



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class TextArea(Component):
    """
    Html Textarea wrapper.
    """
    _aspect_keys = []
    _children = []
    value = Aspect(docstring="Current value of the textarea")
    name = Aspect(docstring="Name of the element for forms.")
    cols = Aspect(docstring="Number of columns.")
    rows = Aspect(docstring="Number of rows.")
    required = Aspect(docstring="Is it required in a form.")
    disabled = Aspect(docstring="Is it disabled ?")
    placeholder = Aspect(docstring="Hint when no value is entered.")
    max_length = Aspect(docstring="Max length of the value.")
    autosize = Aspect(docstring="Auto size the")
    style = Aspect()
    class_name = Aspect()

    def __init__(
            self,
            value: typing.Optional[str] = UNDEFINED,
            name: typing.Optional[str] = UNDEFINED,
            cols: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            rows: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            required: typing.Optional[bool] = UNDEFINED,
            disabled: typing.Optional[bool] = UNDEFINED,
            placeholder: typing.Optional[str] = UNDEFINED,
            max_length: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            autosize: typing.Optional[bool] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            identity: str = None
    ):
        """
        :param value: Current value of the textarea
        :param name: Name of the element for forms.
        :param cols: Number of columns.
        :param rows: Number of rows.
        :param required: Is it required in a form.
        :param disabled: Is it disabled ?
        :param placeholder: Hint when no value is entered.
        :param max_length: Max length of the value.
        :param autosize: Auto size the
        :param style:
        :param class_name:
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Modal(Component):
    """
    A modal overlay the page with a darkened background.

    :CSS:

        ``dazzler-core-modal``
        - ``modal-overlay``
        - ``modal-active``
        - ``modal-content``
        - ``modal-header``
        - ``modal-closer``
        - ``modal-body``
        - ``modal-footer``
    """
    _aspect_keys = []
    _children = []
    body = Aspect(required=True, children=True, docstring="Main content of the modal. (css: modal-body)")  # noqa: E501
    header = Aspect(children=True, docstring="Appear before the body. (css: modal-header)")  # noqa: E501
    footer = Aspect(children=True, docstring="Appear after the body. (css: modal-footer)")  # noqa: E501
    active = Aspect(docstring="Activate the modal.")
    close_button = Aspect(default=True, docstring="Include a close button in the top right corner of the header. (default=True)")  # noqa: E501
    class_name = Aspect(docstring="CSS class of the modal.")
    style = Aspect(docstring="Style object of outer container.")

    def __init__(
            self,
            body: typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]],  # noqa: E501
            header: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            footer: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            active: typing.Optional[bool] = UNDEFINED,
            close_button: typing.Optional[bool] = True,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            identity: str = None
    ):
        """
        :param body: Main content of the modal. (css: modal-body)
        :param header: Appear before the body. (css: modal-header)
        :param footer: Appear after the body. (css: modal-footer)
        :param active: Activate the modal.
        :param close_button: Include a close button in the top right corner
            of the header. (default=True)
        :param class_name: CSS class of the modal.
        :param style: Style object of outer container.
        """
        Component.__init__(self, locals(), identity)



import pathlib
import re
import os
import warnings

# Create a data dir for the requirements once and for all.
import typing

from dazzler.tools import format_tag

from dazzler.errors import (
    InvalidRequirementError, InvalidRequirementKindError
)


class RequirementWarning(UserWarning):
    pass


class Requirement:
    """Represents a static asset to include as dependency for rendering."""
    def __init__(
            self,
            internal: str = None,
            kind: str = None,
            name: str = None,
            package: str = None,
            dev: str = None,
            external: str = None,
            page: str = None
    ):
        """
        :param internal: Local path of the requirement to include on the page.
        :param kind: The file extension.
        :param name: The file name.
        :param package: The package related to this requirement.
        :param dev: Local path of the dev asset to serve instead of internal
            when running in debug mode.
        :param external: URL to serve instead when running with
            ``prefer_external`` mode enable in the configs.
        :param page: The page related to this requirement.
        """
        if internal is None and external is None:
            raise InvalidRequirementError(
                'Please give either an internal or external file'
            )

        self.internal = internal
        self.name = name or os.path.basename(internal or external)
        self.kind = kind or self.name.split(os.extsep)[-1]
        self.dev = dev
        self.external = external
        self.external_only = external and not internal
        self.internal_only = internal and not external

        paths = ['dist', self.name]
        url = ['/dazzler', 'requirements', 'static', 'dist', self.name]
        if package:
            paths.insert(0, package)
            url.insert(3, package)

        if page:
            paths.insert(0, page)
            url.insert(3, page)

        self.internal_static = os.path.join(*paths)
        self.internal_url = '/'.join(url)

        if dev:
            paths[paths.index('dist')] = 'dev'
            dev_name = os.path.basename(dev)
            paths[paths.index(self.name)] = dev_name
            self.dev_static = os.path.join(*paths)
            url[url.index('dist')] = 'dev'
            url[url.index(self.name)] = dev_name
            self.dev_url = '/'.join(url)

        if self.external_only:
            warnings.warn(
                f'No local file for requirement: {self.external}',
                RequirementWarning
            )

    def prepare(self, dev=False, external=False) -> dict:
        attributes = {}
        if self.kind == 'js':
            uri_key = 'src'
        else:
            uri_key = 'href'
            if self.kind == 'css':
                attributes['rel'] = 'stylesheet'
                attributes['type'] = 'text/css'

        if self.external_only:
            uri = self.external
        elif dev and self.dev:
            uri = self.dev_url
        elif external and self.external:
            uri = self.external
            attributes['crossorigin'] = ('{key}', 'crossorigin')
        else:
            uri = self.internal_url

        attributes[uri_key] = uri

        return {
            'kind': self.kind,
            'url': uri,
            'attributes': attributes
        }

    def tag(self, dev=False, external=False):
        prepared = self.prepare(dev, external)
        if self.kind == 'css':
            return format_tag(
                'link', prepared['attributes'], opened=True, close=False
            )
        if self.kind == 'js':
            return format_tag('script', prepared['attributes'])
        if self.kind == 'map':
            return ''
        raise InvalidRequirementKindError(
            f'Invalid requirement kind: {self.kind}'
        )

    def __str__(self):
        return self.internal or self.external

    def __repr__(self):
        return self.tag()


def assets_to_requirements(
        path: str, data: dict,
        dev_data: dict = None,
        dev_path: str = None,
        package_name: str = None,
        external: str = None
) -> typing.List[Requirement]:
    """
    Turns the output of webpack-bundle-tracker into a list of Requirements

    :param path: Path where production assets are located
    :param data: Bundle tracker data.'
    :param dev_data: Bundle tracker dev data, if not provided take data.
    :param dev_path: Path where dev assets are located.
    :param package_name: The name of the package if used for packages.
    :param external: External base path for the requirement.
    :return:
    """
    dev_data = dev_data or data
    dev_path = dev_path or path

    return [
        Requirement(
            internal=os.path.join(path, internal['name']),
            dev=os.path.join(dev_path, dev['name']),
            package=package_name,
            external=f'{external}/{internal["name"]}' if external else None
        )
        for internal, dev in zip(data, dev_data)
    ]


def collect_requirements(directory: str, page: str = None):
    """
    Collect all js/css files in the directory and map them as requirement

    :param directory:
    :param page:
    :return:
    """
    requirements = []

    for current, _, files in os.walk(directory):
        for file in sorted(
                (
                    x for x in files
                    if x.endswith('.js') or x.endswith('.css')
                ),
                key=lambda k: [
                    int(x) if x.isdigit() else x
                    for x in re.split('([0-9]+)', k)
                ]
        ):
            path = os.path.join(current, file)
            requirements.append(
                Requirement(
                    internal=path,
                    page=page,
                    name=str(pathlib.Path(path).relative_to(directory))
                )
            )
    return requirements



import asyncio
import json
import os
import time
import weakref
import uuid
import enum
import base64

from typing import Any, Optional

from itsdangerous import Signer, BadSignature
from aiohttp import web

from ..errors import SessionError
from ._middleware import Middleware
from ._undefined import UNDEFINED


class SessionAction(enum.Enum):
    GET = 1
    SET = 2
    DELETE = 3


class Session:
    """
    Session object available in requests by the middleware

    Access with ``request['session']``. Or from binding: ``context.session``
    """

    def __init__(
            self,
            session_id: str,
            query_queue: asyncio.Queue,
    ):
        """
        :param session_id: The session id to perform operations.
        :param query_queue: To send commands up.
        """
        self.session_id = session_id
        self._query_queue = query_queue

    async def get(self, key: str) -> Any:
        """
        Get an item from the session.

        :param key: The item to fetch.
        :return: The value of the key for the session.
        """
        queue = asyncio.Queue()
        await self._query_queue.put(
            (SessionAction.GET, self.session_id, key, queue)
        )
        return await queue.get()

    async def set(self, key: str, value: Any):
        """
        Associate a value with a key for the session.

        :param key: Key to set for the session
        :param value: The value to set.
        :return:
        """
        await self._query_queue.put(
            (SessionAction.SET, self.session_id, key, value)
        )

    async def delete(self, key):
        """
        Delete the value at key.

        :param key: Key to delete.
        :return:
        """
        await self._query_queue.put(
            (SessionAction.DELETE, self.session_id, key, 0)
        )

    async def pop(self, key):
        """
        Retrieve and delete the key.

        :param key:
        :return:
        """
        data = await self.get(key)
        await self.delete(key)
        return data


class SessionBackEnd:
    """
    Base class to save & load sessions.
    """
    def __init__(self, app):
        """
        :param app: Running dazzler app.
        :type app: dazzler.Dazzler
        """
        self.app = app

    async def set(self, session_id: str, key: str, value: Any):
        """
        Set a key for the session id.

        :param session_id: Session to set data with a key.
        :param key: Key to set
        :param value: Value to keep.
        :return:
        """
        raise NotImplementedError

    async def get(self, session_id: str, key: str):
        """
        Get a key for the session id.

        :param session_id: Session to fetch the data for.
        :param key: Key to get.
        :return:
        """
        raise NotImplementedError

    async def delete(self, session_id: str, key: str):
        """
        Delete a session key value.

        :param session_id: Session to delete the key for.
        :param key: The key to delete.
        :return:
        """
        raise NotImplementedError


class FileSessionBackEnd(SessionBackEnd):
    """
    Session backed by the file system.

    Should only be used for development purpose where redis or another
    solution is not available.
    """
    def __init__(self, app):
        super().__init__(app)
        self.save_directory = os.path.join(
            app.data_dir, 'session-data'
        )
        if not os.path.exists(self.save_directory):
            os.makedirs(self.save_directory)
        app.events.subscribe('dazzler_setup', self.cleanup)
        app.events.subscribe('dazzler_start', self.cleanup)
        app.events.subscribe('dazzler_stop', self.cleanup)

    async def save(self, session_id: str, data: dict):
        await self.acquire(session_id)

        with open(self._session_path(session_id), 'w') as f:
            json.dump(data, f)

        self.release(session_id)

    async def load(self, session_id: str) -> Optional[dict]:
        path = self._session_path(session_id)
        if not os.path.exists(path):
            return UNDEFINED

        await self.acquire(session_id)

        with open(path) as f:
            data = json.load(f)

        os.utime(path, None)

        self.release(session_id)
        return data

    async def acquire(self, session_id: str):
        while self._locked(session_id):
            await asyncio.sleep(0.005)

        self._lock(session_id)

    async def set(self, session_id: str, key: str, value: Any):
        data = await self.load(session_id)

        if data is UNDEFINED:
            data = {}

        data[key] = value
        await self.save(session_id, data)

    async def get(self, session_id: str, key: str) -> Any:
        data = await self.load(session_id)
        if data is not UNDEFINED:
            return data.get(key, UNDEFINED)
        return UNDEFINED

    async def delete(self, session_id: str, key: str):
        data = await self.load(session_id)
        if data is UNDEFINED or key not in data:
            return
        data.pop(key)
        await self.save(session_id, data)

    def release(self, session_id: str):
        os.remove(self._session_path(session_id, lock=True))

    def _locked(self, session_id: str) -> bool:
        return os.path.exists(self._session_path(session_id, lock=True))

    def _session_path(self, session_id: str, lock=False) -> str:
        path = os.path.join(self.save_directory, f'{session_id}.json')
        if lock:
            path += '.lock'
        return path

    def _lock(self, session_id: str):
        with open(self._session_path(session_id, lock=True), 'w') as f:
            f.write('locked')

    async def cleanup(self, _):
        now = time.time()
        for file in os.listdir(self.save_directory):
            if not file.endswith('json'):
                return
            path = os.path.join(self.save_directory, file)
            modified = os.path.getmtime(path)
            if now - modified > self.app.config.session.duration:
                os.remove(path)


class RedisSessionBackend(SessionBackEnd):
    """
    Backed by aioredis.

    Values are serialized to json first to keep the types.

    Install with ``pip install dazzler[redis]``

    :seealso: https://aioredis.readthedocs.io/
    """

    def __init__(self, app):
        super().__init__(app)
        asyncio.get_event_loop().create_task(self._setup())

    async def _setup(self):
        import aioredis
        self.redis = await aioredis.create_redis_pool(
            os.getenv('REDIS_URL', 'redis://localhost:6379')
        )

    async def set(self, session_id: str, key: str, value: Any):
        transaction = self.redis.multi_exec()
        # Serialize to keep the type.
        transaction.hset(session_id, key, json.dumps({'v': value}))
        transaction.expire(
            session_id, self.app.config.session.duration
        )
        await transaction.execute()

    async def get(self, session_id: str, key: str):
        if not await self.redis.hexists(session_id, key):
            return UNDEFINED
        data = await self.redis.hget(session_id, key)
        await self.redis.expire(
            session_id, self.app.config.session.duration
        )
        data = json.loads(data)
        return data['v']

    async def delete(self, session_id: str, key: str):
        await self.redis.hdel(session_id, key)


class SessionMiddleware(Middleware):
    """
    Insert session objects into requests.
    """
    _backend: SessionBackEnd

    def __init__(self, app, backend=None):
        """
        :param app: Dazzler application.
        :type app: dazzler.Dazzler
        :param backend:
        :type backend: SessionBackEnd
        """
        if not app.config.secret_key:
            raise SessionError('Missing app secret key!')
        if app.config.secret_key == 'Please change me':
            app.logger.warning(
                'Please change the app secret key in the configs.'
            )
        self.app = app
        self.signer = Signer(
            app.config.secret_key,
            salt=app.config.session.salt
        )
        self._backend = backend or FileSessionBackEnd(app)
        # pylint: disable=E1121
        self._sessions_queues = weakref.WeakValueDictionary()
        self._query_queue = asyncio.Queue()
        loop = asyncio.get_event_loop()
        self._handlers = [
            loop.create_task(self._handle_queries()),
        ]
        self.app.events.subscribe('dazzler_stop', self._on_stop)

    def verify_session(self, session_id):
        unsigned = self.signer.unsign(session_id).decode()
        session, created = unsigned.split('#')
        return session, int(base64.b64decode(created))

    async def _handle_queries(self):
        while not self.app.stop_event.is_set():
            action, session_id, key, arg = await self._query_queue.get()
            if action == SessionAction.GET:
                data = await self._backend.get(session_id, key)
                await arg.put(data)
            elif action == SessionAction.SET:
                await self._backend.set(session_id, key, arg)
            elif action == SessionAction.DELETE:
                await self._backend.delete(session_id, key)

    def _set_session(self, session_id: str = None):
        session_id = session_id or uuid.uuid4().hex

        created = base64.b64encode(str(int(time.time())).encode()).decode()

        async def set_cookie(response):
            response.set_cookie(
                self.app.config.session.cookie_name,
                self.signer.sign(f'{session_id}#{created}').decode(),
                httponly=True,
                max_age=self.app.config.session.duration,
            )
        return session_id, set_cookie

    async def __call__(self, request: web.Request):
        cookie = request.cookies.get(
            self.app.config.session.cookie_name
        )
        callback = None

        if not cookie:
            session_id, callback = self._set_session()
        else:
            try:
                session_id, created = self.verify_session(cookie)
                delta = time.time() - created
                if delta > self.app.config.session.refresh_after:
                    session_id, callback = self._set_session(session_id)
            except BadSignature as error:
                session_id, callback = self._set_session()
                self.app.logger.exception(error)

        request['session'] = Session(
            session_id,
            self._query_queue,
        )

        return callback

    async def _on_stop(self, _):
        for handler in self._handlers:
            handler.cancel()



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Table(Component):
    """
    Display data in a tabular manner (Non interactive).

    :CSS:
        ``dazzler-core-table``
        - ``table-head``
        - ``table-heading``
        - ``table-body``
        - ``table-row``
        - ``table-cell``
        - ``table-footer``
    """
    _aspect_keys = []
    _children = []
    rows = Aspect(children=True, docstring="Rows of data or components to display.")  # noqa: E501
    caption = Aspect(children=True, docstring="Title of the table")
    headers = Aspect(children=True, docstring="Table headings")
    class_name = Aspect(docstring="CSS class of the table.")
    style = Aspect(docstring="Style of the table element.")
    footer = Aspect(children=True, docstring="Content of tfoot element (.table-footer)")  # noqa: E501
    include_row_number = Aspect(docstring="Each row start with the row number.")  # noqa: E501
    row_number_start = Aspect(docstring="The start of the row number, useful if paged.")  # noqa: E501

    def __init__(
            self,
            rows: typing.Optional[typing.List[typing.List[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]]]] = UNDEFINED,  # noqa: E501
            caption: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            headers: typing.Optional[typing.List[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]]] = UNDEFINED,  # noqa: E501
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            footer: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            include_row_number: typing.Optional[bool] = UNDEFINED,
            row_number_start: typing.Optional[typing.Union[float, int]] = UNDEFINED,  # noqa: E501
            identity: str = None
    ):
        """
        :param rows: Rows of data or components to display.
        :param caption: Title of the table
        :param headers: Table headings
        :param class_name: CSS class of the table.
        :param style: Style of the table element.
        :param footer: Content of tfoot element (.table-footer)
        :param include_row_number: Each row start with the row number.
        :param row_number_start: The start of the row number, useful if
            paged.
        """
        Component.__init__(self, locals(), identity)



class DazzlerError(Exception):
    """Base Dazzler error"""


class PackageConflictError(DazzlerError):
    """A conflict in package name was detected."""


class PageConflictError(DazzlerError):
    """Duplicate page name or url detected."""


class ServerStartedError(DazzlerError):
    """Action require the server to be uninitialized."""


class BindingError(DazzlerError):
    """Error in the binding system"""


class NoInstanceFoundError(DazzlerError):
    """No instance was found when the application is run by command line."""


class GetAspectError(BindingError):
    """Error while retrieving an aspect from the frontend."""


class TriggerLoopError(BindingError):
    """Setting the same aspect that triggered results in infinite loop."""


class RequirementError(DazzlerError):
    """An error in the requirement system"""


class InvalidRequirementError(RequirementError):
    """No internal file or external file was given."""


class InvalidRequirementKindError(RequirementError):
    """Don't know what to do with this kind of requirements."""


class RequirementNotFoundError(RequirementError):
    """The requirement file did not exist."""


class SessionError(DazzlerError):
    """Error related to session system."""


class AuthError(DazzlerError):
    """Error related to authentication system."""



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class DatePicker(Component):
    """
    Pick a date from a calendar.

    :CSS:

        ``dazzler-calendar-date-picker``
        - ``picker``
    """
    _aspect_keys = []
    _children = []
    value = Aspect()
    opened = Aspect()
    date_format = Aspect(default='DD MMM YYYY', docstring="(default='DD MMM YYYY')")  # noqa: E501

    def __init__(
            self,
            value: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            opened: typing.Optional[bool] = UNDEFINED,
            date_format: typing.Optional[str] = 'DD MMM YYYY',
            identity: str = None
    ):
        """
        :param value:
        :param opened:
        :param date_format: (default='DD MMM YYYY')
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Calendar(Component):
    """
    Display a calendar based on the month of a timestamp.

    :CSS:

        ``dazzler-calendar-calendar``
        - ``calendar-header``
        - ``month-label``
        - ``week-labels``
        - ``week-label``
        - ``calendar-day``
        - ``selected-day``
        - ``week-rows``
        - ``calendar-week``
        - ``empty``
    """
    _aspect_keys = []
    _children = []
    month_timestamp = Aspect(docstring="Timestamp of the month to show. If not supplied is today.")  # noqa: E501
    selected = Aspect(docstring="The currently selected day as an object")
    month_format = Aspect(default='MMMM YYYY', docstring="Formatting of the month label (default='MMMM YYYY')")  # noqa: E501
    week_labels = Aspect(default=UNDEFINED, docstring="Labels for the week, starting from sunday. (default=['Su', 'Mo', 'Tu', 'We', 'Th', 'Fr', 'Sa'])")  # noqa: E501
    class_name = Aspect()
    style = Aspect()
    use_selected = Aspect(default=True, docstring="Whether to set selected aspect on click/start up.This will disable highlight of day in the calendar. (default=True)")  # noqa: E501

    def __init__(
            self,
            month_timestamp: typing.Optional[typing.Union[float, int]] = UNDEFINED,  # noqa: E501
            selected: typing.Optional[typing.Dict[str, typing.Union[typing.Union[float, int]]]] = UNDEFINED,  # noqa: E501
            month_format: typing.Optional[str] = 'MMMM YYYY',
            week_labels: typing.Optional[typing.List[str]] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            use_selected: typing.Optional[bool] = True,
            identity: str = None
    ):
        """
        :param month_timestamp: Timestamp of the month to show. If not
            supplied is today.
        :param selected: The currently selected day as an object
        :param month_format: Formatting of the month label (default='MMMM
            YYYY')
        :param week_labels: Labels for the week, starting from sunday.
            (default=['Su', 'Mo', 'Tu', 'We', 'Th', 'Fr', 'Sa'])
        :param class_name:
        :param style:
        :param use_selected: Whether to set selected aspect on click/start
            up. This will disable highlight of day in the calendar.
            (default=True)
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Link(Component):
    """
    Link to external url or other dazzler page by name.

    :CSS: ``dazzler-core-link``
    """
    _aspect_keys = []
    _children = []
    href = Aspect(docstring="The link destination.")
    children = Aspect(children=True, docstring="Text/Component to show as link.")  # noqa: E501
    id = Aspect(docstring="Id of the html element, othewise the identity is used.")  # noqa: E501
    class_name = Aspect(docstring="CSS class of the <a> element")
    style = Aspect(docstring="Style object of root <a> element")
    title = Aspect(docstring="Hovered description")
    page_name = Aspect(docstring="Name of the page to redirect to if the href is not set.")  # noqa: E501

    def __init__(
            self,
            href: typing.Optional[str] = UNDEFINED,
            children: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            id: typing.Optional[str] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            title: typing.Optional[str] = UNDEFINED,
            page_name: typing.Optional[str] = UNDEFINED,
            identity: str = None
    ):
        """
        :param href: The link destination.
        :param children: Text/Component to show as link.
        :param id: Id of the html element, othewise the identity is used.
        :param class_name: CSS class of the <a> element
        :param style: Style object of root <a> element
        :param title: Hovered description
        :param page_name: Name of the page to redirect to if the href is
            not set.
        """
        Component.__init__(self, locals(), identity)



from precept import Config, ConfigProperty, ConfigFormat, Nestable


class DazzlerConfig(Config):
    """Dazzler configuration"""

    app_title = ConfigProperty(
        default='Dazzler',
        comment='Name of the title of the index.'
    )

    host = ConfigProperty(
        default='127.0.0.1',
        comment='Host address',
        auto_global=True,
    )

    port = ConfigProperty(
        default=8150,
        comment='Port of the server',
        auto_global=True,
    )

    debug = ConfigProperty(
        default=False,
        config_type=bool,
        auto_global=True,
    )

    route_prefix = ConfigProperty(
        default='',
        comment='Route prefix for all dazzler related endpoints.',
    )

    static_folder = ConfigProperty(
        default='static',
        comment='Path relative to project folder where files will be served.',
    )

    static_prefix = ConfigProperty(
        default='/static',
        comment='Prefix for the static route'
    )

    secret_key = ConfigProperty(
        default='Please change me',
        comment='Secret key to use for signing sessions.',
        config_type=str,
    )

    class Session(Nestable):
        enable = ConfigProperty(
            config_type=bool,
            default=True
        )
        backend = ConfigProperty(
            comment='Type of session backend to use. Choices: File, Redis',
            default='File',
            config_type=str,
        )
        salt = ConfigProperty(
            config_type=str,
            comment='Salt for signing sessions ids.'
        )
        cookie_name = ConfigProperty(
            config_type=str,
            comment='Name of the session cookie.',
            default='sessionid'
        )
        duration = ConfigProperty(
            config_type=int,
            comment='Maximum duration of a session in seconds. '
                    '(Default=30 days)',
            default=2592000,
        )
        refresh_after = ConfigProperty(
            config_type=int,
            comment='Refresh the session when accessed '
                    'after this number of seconds. (Default=7 days)',
            default=604800
        )

    session: Session

    class Requirements(Nestable):
        prefer_external = ConfigProperty(
            default=False,
            config_type=bool,
            comment='Prefer serving external requirements when available'
        )
        external_scripts = ConfigProperty(
            default=[],
            config_type=list,
            comment='List of urls to include as script requirement.',
        )
        internal_scripts = ConfigProperty(
            default=[],
            config_type=list,
            comment='List of files to include as script requirement.',
        )
        external_styles = ConfigProperty(
            default=[],
            config_type=list,
            comment='List of urls to include as style requirement.'
        )
        internal_styles = ConfigProperty(
            default=[],
            config_type=list,
            comment='List of files to include as script requirement.',
        )

    requirements: Requirements

    class Renderer(Nestable):
        retries = ConfigProperty(
            default=20,
            config_type=int,
            comment='Number of times it will try to reconnect '
                    'when the websocket connection is lost.'
        )
        ping = ConfigProperty(
            default=False,
            config_type=bool,
            comment='Enable to send a ping every interval to keep the '
                    'websocket connected if it didn\'t send data after '
                    'a delay. Some hosts providers will automatically closes'
                    'idling connection after a while.'
        )
        ping_interval = ConfigProperty(
            default=25.0,
            config_type=float,
            comment='Interval at which to send ping data.'
        )

    renderer: Renderer

    class Authentication(Nestable):
        enable = ConfigProperty(
            default=False,
            config_type=bool,
        )

        authenticator = ConfigProperty(
            config_type=str,
            comment='Path to an instance or subclass of '
                    '`dazzler.system.auth.Authenticator',
            default=''
        )

        backend = ConfigProperty(
            config_type=str,
            comment='Path to an instance or subclass of '
                    '`dazzler.system.auth.AuthBackend',
            default='',
        )

    authentication: Authentication

    class Development(Nestable):
        reload = ConfigProperty(
            default=False,
            config_type=bool,
            auto_global=True,
            global_name='reload',
            comment='Enable hot reload when files used by the application'
                    ' are changed.'
        )

        reload_interval = ConfigProperty(
            default=0.5,
            config_type=float,
            comment='Interval at which the reload checks for file changes.'
        )

        reload_threshold = ConfigProperty(
            default=3.0,
            config_type=float,
            comment='Time to wait from first detected change '
                    'to actual reload.'
        )

    development: Development

    def __init__(self):
        super().__init__(root_name='dazzler', config_format=ConfigFormat.TOML)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Select(Component):
    """
    A select
    """
    _aspect_keys = []
    _children = []
    options = Aspect(required=True, docstring="Option of the select.")
    multi = Aspect(docstring="Controls wether multiple options can be selected.")  # noqa: E501
    value = Aspect(docstring="Value of the selected option(s)")
    size = Aspect(docstring="Size of the options to use instead of")
    class_name = Aspect(docstring="CSS class of the Select")
    placeholder = Aspect(docstring="Placeholder text when the input has no value.")  # noqa: E501
    name = Aspect(docstring="name of the html input that will be created with the current value")  # noqa: E501
    disabled = Aspect(docstring="Disable the component.")

    def __init__(
            self,
            options: typing.List[typing.Dict[str, typing.Union[typing.Union[typing.Union[float, int], str], str]]],  # noqa: E501
            multi: typing.Optional[bool] = UNDEFINED,
            value: typing.Optional[typing.Union[typing.Union[float, int], str, typing.List[typing.Union[typing.Union[float, int], str]]]] = UNDEFINED,  # noqa: E501
            size: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            placeholder: typing.Optional[str] = UNDEFINED,
            name: typing.Optional[str] = UNDEFINED,
            disabled: typing.Optional[bool] = UNDEFINED,
            identity: str = None
    ):
        """
        :param options: Option of the select.
        :param multi: Controls wether multiple options can be selected.
        :param value: Value of the selected option(s)
        :param size: Size of the options to use instead of
        :param class_name: CSS class of the Select
        :param placeholder: Placeholder text when the input has no value.
        :param name: name of the html input that will be created with the
            current value
        :param disabled: Disable the component.
        """
        Component.__init__(self, locals(), identity)



"""
Dazzler systems (API)

- Requirements are JS/CSS resources to include on rendering.
- Packages hold components info and it's requirements.
- Component Aspects are shared between backend and frontend with bindings.
- Generate components with ``dazzler generate metadata.json output_dir``
- Page holds meta data for rendering, requirements, routes, layout, bindings.
"""
from ._component import Component, Aspect  # noqa: F401
from ._binding import *  # noqa: F401, F403
from ._package import Package  # noqa: F401
from ._requirements import (  # noqa: F401
    Requirement,
    RequirementWarning,
    assets_to_requirements,
    collect_requirements,
)
from ._generator import generate_components, generate_meta  # noqa: F401
from ._undefined import UNDEFINED, Undefined  # noqa: F401
from ._page import Page  # noqa: F401
from ._middleware import Middleware  # noqa: F401
from ._route import Route, RouteMethod  # noqa: F401


__all__ = [  # noqa: F405
    'Component',
    'Aspect',
    'BindingContext',
    'Binding',
    'Trigger',
    'State',
    'BoundAspect',
    'Package',
    'Requirement',
    'RequirementWarning',
    'assets_to_requirements',
    'collect_requirements',
    'generate_components',
    'generate_meta',
    'UNDEFINED',
    'Undefined',
    'Page',
    'Middleware',
    'Route',
    'RouteMethod',
]



"""Aspects Binding"""
import functools
import typing
import uuid
import asyncio

from aiohttp import web

from .session import Session
from ._undefined import UNDEFINED
from ._component import prepare_aspects
from ._package import Package


from ..errors import TriggerLoopError, GetAspectError


__all__ = [
    'Trigger', 'State', 'BoundAspect', 'Binding', 'BindingContext',
]


def is_component(aspect):
    return isinstance(aspect, dict) \
           and 'identity' in aspect \
           and 'aspects' in aspect \
           and 'name' in aspect \
           and 'package' in aspect


def hydrate(aspect):
    if is_component(aspect):
        pack = Package.package_registry[aspect['package']]
        cls = pack.components[aspect['name']]
        return cls(**hydrate(aspect['aspects']), identity=aspect['identity'])
    if isinstance(aspect, dict):
        data = {}
        for key, value in aspect.items():
            data[key] = hydrate(value)
        return data
    if isinstance(aspect, list):
        return [hydrate(x) for x in aspect]
    return aspect


class BoundValue(typing.NamedTuple):
    identity: str
    aspect: str
    value: typing.Any


class _Bind:
    def __init__(self, identity: str, aspect: str, regex=False):
        """
        :param identity: The identity of the component to bind.
        :param aspect: Aspect name to trigger/state.
        :param regex: Identity and aspects are converted into regex and
            matched against for trigger and states.
        """
        self.identity = identity
        self.aspect = aspect
        self.regex = regex

    def __str__(self):
        return f'{self.aspect}@{self.identity}'

    def __repr__(self):
        return f'<{self.__class__.__name__} {self}>'

    def __eq__(self, other):
        return isinstance(other, _Bind) and str(self) == str(other)

    def __hash__(self):
        return hash(str(self))

    def prepare(self) -> dict:
        """
        Prepare the trigger or state for serialization.

        :return: Identity and aspect in a dict.
        """
        return {
            'identity': self.identity,
            'aspect': self.aspect,
            'regex': self.regex,
        }


# pylint: disable=too-few-public-methods
class Trigger(_Bind):
    """Trigger the bound aspect callbacks."""


TriggerList = typing.List[Trigger]


# pylint: disable=too-few-public-methods
class State(_Bind):
    """Usable aspect value in bound aspect without trigger on change."""


StateList = typing.List[State]


class BoundAspect:
    """
    A trigger aspect bound to a method intended to be called when changed.

    Data holder for the handler, trigger and state.
    """
    def __init__(
            self,
            handler,
            trigger: typing.Union[TriggerList, Trigger],
            states: StateList = None,
    ):
        self.handler = handler
        self.trigger = trigger
        self.states = states or []

    def prepare(self) -> list:
        """
        Prepare the binding for serialization.

        :return: list of dict with trigger and states definitions.
        """
        return [
            {
                'trigger': trigger.prepare(),
                'states': [state.prepare() for state in self.states],
                'key': str(trigger),
                'regex': trigger.regex
            } for trigger in self.triggers
        ]

    @property
    def triggers(self):
        if isinstance(self.trigger, Trigger):
            return [self.trigger]
        return self.trigger

    async def __call__(self, *args, **kwargs):
        return await self.handler(*args, **kwargs)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        # Patch self if used in a class.
        return functools.partial(self.__call__, instance)

    def __str__(self):
        return str(self.trigger)

    def __repr__(self):
        return f'<BoundAspect {self}>'

    def __eq__(self, other):
        return isinstance(other, BoundAspect) and str(self) == str(other)

    def __hash__(self):
        return hash(str(self))


class BindingContext:
    """The context in which the bound function execute."""
    def __init__(
            self,
            identity: str,
            request: web.Request,
            trigger: BoundValue,
            states: typing.Dict[str, BoundValue],
            websocket: web.WebSocketResponse,
            request_queue: asyncio.Queue,
    ):
        self.identity = identity
        self.request = request
        self.trigger = trigger
        self.states = states
        self.websocket = websocket
        self.session: Session = request.get('session')
        self._request_queue = request_queue
        self._response_queue = asyncio.Queue()

    async def set_aspect(self, identity, **aspects):
        """
        Update aspects of a component on the front end.

        :param identity: Identity of the component to update.
        :param aspects: The aspects to set on the component.
        :return:
        """
        if identity == self.identity:
            if any(x == self.trigger.aspect for x in aspects):
                trigger_key = f'{self.trigger.identity}.{self.trigger.aspect}'
                raise TriggerLoopError(
                    f'Setting the same aspect that triggered: {trigger_key}'
                )

        regex = isinstance(identity, typing.Pattern)

        if regex:
            identity = identity.pattern

        await self.websocket.send_json({
            'kind': 'set-aspect',
            'identity': str(identity),
            'regex': regex,
            'payload': prepare_aspects(aspects)
        })

    async def get_aspect(self, identity: str, aspect: str):
        """
        Request the value of an aspect from the frontend.

        :param identity: Component to get aspect from.
        :param aspect: Name of the aspect property.
        :return:
        """

        response_queue = asyncio.Queue()

        await self._request_queue.put({
            'request_id': uuid.uuid4().hex,
            'queue': response_queue,
            'identity': identity,
            'aspect': aspect,
            'kind': 'get-aspect'
        })

        value, error = await response_queue.get()
        if value is UNDEFINED:
            raise GetAspectError(
                error or f'Undefined aspect {aspect}@{identity}'
            )

        return hydrate(value)

    async def _get_storage(self, storage, identity):
        response_queue = asyncio.Queue()

        await self._request_queue.put({
            'request_id': uuid.uuid4().hex,
            'queue': response_queue,
            'identity': identity,
            'storage': storage,
            'kind': 'get-storage'
        })

        value, error = await response_queue.get()
        if error is not UNDEFINED:
            raise error

        return value

    async def _set_storage(self, storage, identity, payload):
        await self._request_queue.put({
            'request_id': uuid.uuid4().hex,
            'identity': identity,
            'kind': 'set-storage',
            'storage': storage,
            'payload': payload
        })

    async def get_local_storage(self, identity):
        return await self._get_storage('local', identity)

    async def set_local_storage(self, identity, payload):
        await self._set_storage('local', identity, payload)

    async def get_session_storage(self, identity):
        return await self._get_storage('session', identity)

    async def set_session_storage(self, identity, payload):
        await self._set_storage('session', identity, payload)


# Decorator
class Binding:
    """Bind a function to execute when the trigger aspect change."""
    def __init__(
            self,
            trigger: Trigger,
            states: StateList = None,
    ):
        self.trigger = trigger
        self.states = states or []

    def __call__(self, func):

        @functools.wraps(func)
        async def bound(request, data, ws, request_queue):
            trigger = BoundValue(
                data['trigger']['identity'],
                data['trigger']['aspect'],
                hydrate(data['trigger']['value'])
            )
            states = {}
            for state in data['states']:
                component = states.setdefault(state['identity'], {})
                component[state['aspect']] = hydrate(
                    state.get('value', UNDEFINED)
                )

            context = BindingContext(
                trigger.identity,
                request,
                trigger,
                states,
                ws,
                request_queue
            )
            return await func(context)

        return BoundAspect(bound, self.trigger, self.states)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class ListBox(Component):
    """
    A component where you can ``add`` items to instead of rendering
    the whole list again. With an optional max size.

    :CSS:

         ``dazzler-core-list-box``

    :example:

    .. literalinclude:: ../../tests/components/pages/list_box.py
    """
    _aspect_keys = []
    _children = []
    items = Aspect(default=UNDEFINED, children=True, docstring="List of items to render. (default=[])")  # noqa: E501
    max_length = Aspect(docstring="Maximum amount of items in the list, extra items will be popped off.")  # noqa: E501
    append = Aspect(children=True, docstring="Add an item to the end of the list.")  # noqa: E501
    prepend = Aspect(children=True, docstring="Add an item to the start of the list.")  # noqa: E501
    concat = Aspect(children=True, docstring="Concatenate another list with the current items.")  # noqa: E501
    insert = Aspect(children=True, docstring="Insert an item at an index position.")  # noqa: E501
    delete_index = Aspect(docstring="Delete the item at the index.")
    scrollable = Aspect(docstring="Wether the list box container is scrollable.")  # noqa: E501
    direction = Aspect(default="'vertical'", docstring="In which direction the items will be inserted. (Possible values: 'vertical', 'horizontal') (default="'vertical'")")  # noqa: E501
    size = Aspect(docstring="Must be set for the scrollable aspect to work.Height or width in pixels depending on the direction aspect.")  # noqa: E501
    keep_scroll = Aspect(docstring="Keep the last appended item in view if scrolling is enabled.")  # noqa: E501
    class_name = Aspect()
    style = Aspect()

    def __init__(
            self,
            items: typing.Optional[typing.List[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]]] = UNDEFINED,  # noqa: E501
            max_length: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            append: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            prepend: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            concat: typing.Optional[typing.List[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]]] = UNDEFINED,  # noqa: E501
            insert: typing.Optional[typing.Dict[str, typing.Union[typing.Union[float, int], typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]]]] = UNDEFINED,  # noqa: E501
            delete_index: typing.Optional[typing.Union[float, int]] = UNDEFINED,  # noqa: E501
            scrollable: typing.Optional[bool] = UNDEFINED,
            direction: typing.Optional[typing.Any] = "'vertical'",
            size: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            keep_scroll: typing.Optional[bool] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            identity: str = None
    ):
        """
        :param items: List of items to render. (default=[])
        :param max_length: Maximum amount of items in the list, extra items
            will be popped off.
        :param append: Add an item to the end of the list.
        :param prepend: Add an item to the start of the list.
        :param concat: Concatenate another list with the current items.
        :param insert: Insert an item at an index position.
        :param delete_index: Delete the item at the index.
        :param scrollable: Wether the list box container is scrollable.
        :param direction: In which direction the items will be inserted.
            (Possible values: 'vertical', 'horizontal') (default="'vertical'")
        :param size: Must be set for the scrollable aspect to work. Height
            or width in pixels depending on the direction aspect.
        :param keep_scroll: Keep the last appended item in view if
            scrolling is enabled.
        :param class_name:
        :param style:
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Pager(Component):
    """
    Paging for dazzler apps.
    """
    _aspect_keys = []
    _children = []
    total_items = Aspect(required=True, docstring="The total items in the set.")  # noqa: E501
    items_per_page = Aspect(default=10, docstring="The number of items a page contains. (default=10)")  # noqa: E501
    style = Aspect()
    class_name = Aspect()
    page_style = Aspect(docstring="Style for the page numbers.")
    page_class_name = Aspect(docstring="CSS class for the page numbers.")
    pages_displayed = Aspect(default=10, docstring="The number of pages displayed by the pager. (default=10)")  # noqa: E501
    pages = Aspect(docstring="Read only, the currently displayed pages numbers.")  # noqa: E501
    current_page = Aspect(default=1, docstring="The current selected page. (default=1)")  # noqa: E501
    total_pages = Aspect(docstring="Set by total_items / items_per_page")
    start_offset = Aspect(docstring="The starting index of the current pageCan be used to slice data eg: data[start_offset: end_offset]")  # noqa: E501
    end_offset = Aspect(docstring="The end index of the current page.")

    def __init__(
            self,
            total_items: typing.Union[float, int],
            items_per_page: typing.Optional[typing.Union[float, int]] = 10,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            page_style: typing.Optional[typing.Dict] = UNDEFINED,
            page_class_name: typing.Optional[str] = UNDEFINED,
            pages_displayed: typing.Optional[typing.Union[float, int]] = 10,
            pages: typing.Optional[typing.List] = UNDEFINED,
            current_page: typing.Optional[typing.Union[float, int]] = 1,
            total_pages: typing.Optional[typing.Union[float, int]] = UNDEFINED,  # noqa: E501
            start_offset: typing.Optional[typing.Union[float, int]] = UNDEFINED,  # noqa: E501
            end_offset: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            identity: str = None
    ):
        """
        :param total_items: The total items in the set.
        :param items_per_page: The number of items a page contains.
            (default=10)
        :param style:
        :param class_name:
        :param page_style: Style for the page numbers.
        :param page_class_name: CSS class for the page numbers.
        :param pages_displayed: The number of pages displayed by the pager.
            (default=10)
        :param pages: Read only, the currently displayed pages numbers.
        :param current_page: The current selected page. (default=1)
        :param total_pages: Set by total_items / items_per_page
        :param start_offset: The starting index of the current page Can be
            used to slice data eg: data[start_offset: end_offset]
        :param end_offset: The end index of the current page.
        """
        Component.__init__(self, locals(), identity)



import asyncio
import functools
from typing import Optional, Callable, Awaitable
from urllib.parse import quote
from aiohttp import web

from dazzler.system import UNDEFINED
from ._middleware import Middleware
from ._page import Page


def _default_page(default_redirect):
    from dazzler.components import core, auth

    async def layout(request: web.Request):
        next_url = request.query.get('next_url') or default_redirect
        error = request.query.get('err')

        footer = core.Container(
            'Invalid credentials',
            style={'color': 'red', 'padding': '0.5rem'},
            identity='login-error'
        ) if error else UNDEFINED

        return core.Container([
            auth.Login(
                '/auth/login',
                method='post',
                identity='login-form',
                next_url=next_url,
                bordered=True,
                header=core.Html('h2', 'Please sign in'),
                footer=footer,
                style={
                    'padding': '2rem',
                }
            ),
        ], style={
            'display': 'flex',
            'justifyContent': 'center',
            'alignItems': 'center',
            'width': '100%',
            'marginTop': '4rem'
        })

    page = Page(__name__, layout, packages=['dazzler_core', 'dazzler_auth'])

    return page


class User:
    """
    Base user of the dazzler auth system.
    """
    username: str

    def __init__(self, username: str):
        self.username = username


class AuthBackend:
    """
    Handle the request part of authentication protocols.
    """

    async def is_authenticated(self, request: web.Request) -> bool:
        """
        If the request is authenticated.

        :param request: The request to verify if authentication credentials
            are provided.
        :return: Whether the user is authenticated.
        """
        raise NotImplementedError

    async def login(
            self, user: User, request: web.Request, response: web.Response
    ):
        """
        Do what it takes to store the login info after being successfully
        authenticated by the authenticator.

        :param user: The authenticated user
        :param request: Incoming request.
        :param response: A redirect response if needed to set cookies.
        :return:
        """
        raise NotImplementedError

    async def logout(
            self, user: User, request: web.Request, response: web.Response
    ):
        """
        Delete what makes this user authenticated in the system.

        :param user: User to logout.
        :param request: Incoming request.
        :param response: A redirect response if need to clear cookies.
        :return:
        """
        raise NotImplementedError

    async def get_username(self, request: web.Request) -> str:
        """
        Get the username from the request.

        :param request: Incoming request.
        :return: Retrieved username.
        """
        raise NotImplementedError


class AuthSessionBackend(AuthBackend):
    async def is_authenticated(self, request: web.Request) -> bool:
        username = await request['session'].get('username')
        return username is not UNDEFINED

    async def login(
            self, user: User,
            request: web.Request,
            response: web.Response
    ):
        session = request['session']
        await session.set('username', user.username)

    async def logout(self, user, request, response):
        session = request['session']
        await session.delete('username')

    async def get_username(self, request: web.Request) -> str:
        return await request['session'].get('username')


class Authenticator:
    """
    Base authenticator, a subclass of this must be presented to
    DazzlerAuth init in order to provide authentication for an app.
    """

    async def authenticate(self, username: str, password: str) \
            -> Optional[User]:
        """
        For login purpose.

        :param username:
        :param password:
        :return:
        """
        raise NotImplementedError

    async def get_user(self, username: str) -> User:
        """
        Retrieve a user by it's username.

        :param username:
        :return:
        """
        raise NotImplementedError


class AuthMiddleware(Middleware):
    """
    Add the user if authenticated to the request object.
    """

    def __init__(self, app, auth):
        self.app = app
        self.auth = auth

    async def __call__(
            self, request: web.Request
    ) -> Optional[Callable[[web.Response], Awaitable]]:
        if await self.auth.backend.is_authenticated(request):
            request['user'] = await self.auth.authenticator.get_user(
                (await self.auth.backend.get_username(request))
            )
        return None


class DazzlerAuth:
    """
    Handle the logic for page authentication.

    Requires an authenticator to provide

    Default to session backend, make sure a session middleware is present.
    """

    def __init__(
            self, app,
            authenticator: Authenticator,
            backend: AuthBackend = None,
            login_page: Page = None,
            default_redirect: str = None,
    ):
        """
        :param app:
        :type app: dazzler.Dazzler
        :param backend:
        """
        self.app = app
        self.backend = backend or AuthSessionBackend()
        self.authenticator = authenticator
        self.app.middlewares.append(AuthMiddleware(app, self))
        self.login_page = login_page or _default_page(default_redirect)
        app.server.route_page = self.require_page_login(app.server.route_page)
        app.server.route_page_json = self.require_page_login(
            app.server.route_page_json, redirect=False,
        )
        app.server.route_update = self.require_page_login(
            app.server.route_update, redirect=False,
        )

        self.login_page.route('/login', method='post')(self.login)
        self.login_page.route('/logout', method='post')(self.logout)
        app.add_page(self.login_page)

    async def login(self, request: web.Request):
        data = await request.post()
        next_url = data.get('next_url')
        username = data.get('username')
        password = data.get('password')
        response = web.HTTPSeeOther(location=next_url)
        user = await self.authenticator.authenticate(username, password)
        if user:
            await self.backend.login(user, request, response)
            raise response

        # Cheap throttling on failures.
        await asyncio.sleep(0.25)
        login_url = request.app.router[self.login_page.name].url_for()
        if self.login_page:
            raise web.HTTPFound(
                location=f'{login_url}?next_url={quote(next_url)}&err=1'
            )
        raise web.HTTPUnauthorized()

    async def logout(self, request: web.Request):
        data = await request.post()

        next_url = data.get('next_url')
        response = web.HTTPSeeOther(location=next_url)
        user = await self.authenticator.get_user(
            (await self.backend.get_username(request))
        )
        await self.backend.logout(user, request, response)
        raise response

    def require_page_login(self, func, redirect=True):

        @functools.wraps(func)
        async def auth_page_wrapper(request: web.Request, page: Page):
            if page.require_login:
                if not await self.backend.is_authenticated(request):
                    if self.login_page and redirect:
                        url = str(
                            request.app.router[self.login_page.name].url_for()
                        )
                        next_url = str(request.url)

                        raise web.HTTPFound(
                            location=f'{url}?next_url={quote(next_url)}'
                        )
                    raise web.HTTPUnauthorized()
            return await func(request, page)

        return auth_page_wrapper



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class ProgressBar(Component):
    """
    Simple progress bar support by all browsers.

    :CSS:

        ``dazzler-core-progress-bar``
        - ``progress``
        - ``high``
        - ``low``
        - ``optimum``
        - ``striped``
        - ``rounded``
    """
    _aspect_keys = []
    _children = []
    value = Aspect(default=0, docstring="The current numeric value. This must be between the minimum and maximum values (min attribute and max attribute) if they are specified. If unspecified or malformed, the value is 0. If specified, but not within the range given by the min attribute and max attribute, the value is equal to the nearest end of the range. (default=0)")  # noqa: E501
    minimum = Aspect(docstring="The lower numeric bound of the measured range. This must be less than the maximum value (max attribute), if specified. If unspecified, the minimum value is 0")  # noqa: E501
    maximum = Aspect(docstring="The upper numeric bound of the measured range. This must be greater than the minimum value (min attribute), if specified. If unspecified, the maximum value is 1.")  # noqa: E501
    low = Aspect(docstring="The upper numeric bound of the low end of the measured range. This must be greater than the minimum value (min attribute), and it also must be less than the high value and maximum value (high attribute and max attribute, respectively), if any are specified. If unspecified, or if less than the minimum value, the low value is equal to the minimum value.")  # noqa: E501
    high = Aspect(docstring="The lower numeric bound of the high end of the measured range. This must be less than the maximum value (max attribute), and it also must be greater than the low value and minimum value (low attribute and min attribute, respectively), if any are specified. If unspecified, or if greater than the maximum value, the high value is equal to the maximum value.")  # noqa: E501
    optimum = Aspect(docstring="This attribute indicates the optimal numeric value. It must be within the range (as defined by the min attribute and max attribute). When used with the low attribute and high attribute, it gives an indication where along the range is considered preferable. For example, if it is between the min attribute and the low attribute, then the lower range is considered preferred.")  # noqa: E501
    id = Aspect(docstring="Id for this component, otherwise the identity is used.")  # noqa: E501
    class_name = Aspect(docstring="CSS class of the outer container.")
    progress_class_name = Aspect(docstring="CSS class of the inner progress container.")  # noqa: E501
    style = Aspect(docstring="Style of the component.")
    progress_text = Aspect(docstring="Set to show the value on the progress bar. (Possible values: 'percent', 'value')")  # noqa: E501
    click = Aspect(docstring="Readonly click event with the value.")
    striped = Aspect(docstring="Use striped style.")
    rounded = Aspect(docstring="Use the rounded style")

    def __init__(
            self,
            value: typing.Optional[typing.Union[float, int]] = 0,
            minimum: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            maximum: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            low: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            high: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            optimum: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            id: typing.Optional[str] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            progress_class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            progress_text: typing.Optional[typing.Any] = UNDEFINED,
            click: typing.Optional[typing.Dict[str, typing.Union[typing.Union[float, int]]]] = UNDEFINED,  # noqa: E501
            striped: typing.Optional[bool] = UNDEFINED,
            rounded: typing.Optional[bool] = UNDEFINED,
            identity: str = None
    ):
        """
        :param value: The current numeric value. This must be between the
            minimum and maximum values (min attribute and max attribute) if
            they are specified. If unspecified or malformed, the value is 0. If
            specified, but not within the range given by the min attribute and
            max attribute, the value is equal to the nearest end of the range.
            (default=0)
        :param minimum: The lower numeric bound of the measured range. This
            must be less than the maximum value (max attribute), if specified.
            If unspecified, the minimum value is 0
        :param maximum: The upper numeric bound of the measured range. This
            must be greater than the minimum value (min attribute), if
            specified. If unspecified, the maximum value is 1.
        :param low: The upper numeric bound of the low end of the measured
            range. This must be greater than the minimum value (min attribute),
            and it also must be less than the high value and maximum value
            (high attribute and max attribute, respectively), if any are
            specified. If unspecified, or if less than the minimum value, the
            low value is equal to the minimum value.
        :param high: The lower numeric bound of the high end of the
            measured range. This must be less than the maximum value (max
            attribute), and it also must be greater than the low value and
            minimum value (low attribute and min attribute, respectively), if
            any are specified. If unspecified, or if greater than the maximum
            value, the high value is equal to the maximum value.
        :param optimum: This attribute indicates the optimal numeric value.
            It must be within the range (as defined by the min attribute and
            max attribute). When used with the low attribute and high
            attribute, it gives an indication where along the range is
            considered preferable. For example, if it is between the min
            attribute and the low attribute, then the lower range is considered
            preferred.
        :param id: Id for this component, otherwise the identity is used.
        :param class_name: CSS class of the outer container.
        :param progress_class_name: CSS class of the inner progress
            container.
        :param style: Style of the component.
        :param progress_text: Set to show the value on the progress bar.
            (Possible values: 'percent', 'value')
        :param click: Readonly click event with the value.
        :param striped: Use striped style.
        :param rounded: Use the rounded style
        """
        Component.__init__(self, locals(), identity)



from ._button import Button
from ._check_list import CheckList
from ._container import Container
from ._data_list import DataList
from ._dropdown import Dropdown
from ._form import Form
from ._grid import Grid
from ._html import Html
from ._image import Image
from ._input import Input
from ._interval import Interval
from ._link import Link
from ._list_box import ListBox
from ._modal import Modal
from ._progress_bar import ProgressBar
from ._radio_list import RadioList
from ._select import Select
from ._slider import Slider
from ._store import Store
from ._table import Table
from ._text_area import TextArea
from ._view_port import ViewPort

__all__ = [
    "Button",
    "CheckList",
    "Container",
    "DataList",
    "Dropdown",
    "Form",
    "Grid",
    "Html",
    "Image",
    "Input",
    "Interval",
    "Link",
    "ListBox",
    "Modal",
    "ProgressBar",
    "RadioList",
    "Select",
    "Slider",
    "Store",
    "Table",
    "TextArea",
    "ViewPort"
]



import asyncio
import json
import keyword
import os
import sys
import textwrap
import re
import shlex

import stringcase

from ..tools import OrderedSet, replace_all
from .._assets import meta_path


def _default_prop_type(_):
    return 'typing.Any'


def _optional(value):
    return f'typing.Optional[{value}]'


def _json_or_undefined(obj):
    return json.dumps(obj) if obj else 'UNDEFINED'


def _default_prop_docstring_format(prop, defaults):
    value = defaults.get(prop)
    if not value:
        return ''
    return '(default={})'.format(value)


FILTERED_PROPS = (
    'updateAspects',
    'identity',
)

FILTERED_PROP_TYPES = (
    'func',
    'custom',
    'symbol',
    'instanceOf'
)

UNSUPPORTED_DEFAULT = (
    'array', 'arrayOf', 'object', 'objectOf', 'shape'
)


# React PropTypes to python typing
PROP_TYPING = {
    'array': lambda t: 'typing.List',
    'arrayOf': lambda t: 'typing.List[{}]'.format(
        PROP_TYPING.get(  # type: ignore
            t['value']['name'],
            _default_prop_type
        )(t['value'])
    ),
    'object': lambda t: 'typing.Dict',
    'shape': lambda t: 'typing.Dict[str, typing.Union[{}]]'.format(
        ', '.join(OrderedSet(*(PROP_TYPING.get(  # type: ignore
            item['name'],
            _default_prop_type
        )(item) for item in t['value'].values())))
    ),
    'string': lambda t: 'str',
    'bool': lambda t: 'bool',
    'number': lambda t: 'typing.Union[float, int]',
    'node': lambda t: 'typing.Union[str, int, float, Component,'
                      ' typing.List[typing.Union'
                      '[str, int, float, Component]]]',
    'func': lambda t: 'None',
    'element': lambda t: 'Component',
    # oneOfType get serialized to union...
    'union': lambda t: 'typing.Union[{}]'.format(
        ', '.join(OrderedSet(*(
            PROP_TYPING.get(  # type: ignore
                x['name'],
                _default_prop_type
            )(x)
            for x in t['value']
        )))
    ),
    'any': _default_prop_type,
    'custom': _default_prop_type,
    # There is no proper typing support for enum.
    # It does a strange serialization (not json)
    # The possible values get formatted in the docstring.
    'enum': _default_prop_type,
    'objectOf': lambda t: 'typing.Dict[str, {}]'.format(
        PROP_TYPING.get(  # type: ignore
            t['value']['name'],
            _default_prop_type
        )(t['value'])
    )
}

DEFAULT_PROP_MAPPING = {
    'array': _json_or_undefined,
    'arrayOf': _json_or_undefined,
    'object': _json_or_undefined,
    'shape': _json_or_undefined,
    'string': lambda t: t,
    'bool': lambda t: str(str(t).lower() == 'true'),
    'number': str,
    'node': lambda t: 'UNDEFINED',
    'func': lambda t: 'UNDEFINED',
    'element': lambda t: 'UNDEFINED',
    'oneOf': str,
    'union': str,
    'any': str
}

COMPONENT_TYPENAME = (
    'node',
    'element'
)

POSSIBLE_COMPONENT_TYPENAME = COMPONENT_TYPENAME + (
    'arrayOf',
    'shape',
    'objectOf',
    'union'
)

TEMPLATE = '''
"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class %(name)(Component):
    """
%(docstring)
    """
    _aspect_keys = []
    _children = []
%(aspects)
    def __init__(
            self,
            %(aspects_init)
    ):
        """
        %(init_docstring)
        """
        Component.__init__(self, locals(), identity)
'''.strip()


def is_component_aspect(type_obj):
    type_name = type_obj['name']

    if type_name not in POSSIBLE_COMPONENT_TYPENAME:
        return False

    if type_name in COMPONENT_TYPENAME:
        return True

    if type_name in ('arrayOf', 'objectOf'):
        return is_component_aspect(type_obj['value'])
    if type_name == 'shape':
        return any(is_component_aspect(t) for t in type_obj['value'].values())
    if type_name == 'union':
        return any(is_component_aspect(t) for t in type_obj['value'])
    return False


# pylint: disable=too-many-locals, too-many-statements
def generate_component(display_name, description, props, output_path):
    aspects = []
    aspects_required = []
    aspects_optional = []
    init_docstring = []

    filtered_props = {
        k: v for k, v in props.items()
        if k not in FILTERED_PROPS
        and not k.endswith('*')
        and k not in keyword.kwlist
        and v.get('type', {}).get('name') not in FILTERED_PROP_TYPES
    }

    for name, prop in filtered_props.items():
        aspect_args = []
        required = prop.get('required')
        type_info = prop.get('type')
        if not type_info:
            print(f'Invalid prop: {display_name}.{name}', file=sys.stderr)
            continue
        type_name = type_info.get('name')
        default = prop.get('defaultValue')
        prop_description = prop.get('description', '')
        computed = default and default.get('computed')

        typed = PROP_TYPING.get(
            type_name,
            _default_prop_type
        )(type_info)

        docstring = ' ' + prop_description if prop_description else ''
        if type_name == 'enum':
            docstring += ' (Possible values: {})'.format(
                ', '.join(
                    x.get('value') for x in type_info.get('value')
                )
            )

        if required:
            init_value = f'{name}: {typed},'
            if len(init_value) > 67:
                init_value += '  # noqa: E501'
            aspects_required.append(init_value)
            aspect_args.append('required=True')
        else:
            if not default or computed or type_name in UNSUPPORTED_DEFAULT:
                mapped_default = 'UNDEFINED'
                if default and not computed:
                    doc_default = DEFAULT_PROP_MAPPING.get(
                        type_name,
                        json.dumps
                    )(default.get('value'))
                else:
                    doc_default = 'UNDEFINED'
            else:
                mapped_default = DEFAULT_PROP_MAPPING.get(
                    type_name,
                    json.dumps
                )(default.get('value'))
                doc_default = mapped_default
            if default and not computed:
                aspect_args.append(f'default={mapped_default}')
            if doc_default != 'UNDEFINED':
                if type_name in UNSUPPORTED_DEFAULT:
                    doc_default = doc_default.strip('"')
                docstring += f' (default={doc_default})'
            init_value = f'{name}: {_optional(typed)} = {mapped_default},'
            if len(init_value) > 66:
                init_value += '  # noqa: E501'
            aspects_optional.append(init_value)

        init_docstring.append(f'{os.linesep}            '.join(
            textwrap.fill(
                f':param {name}:{docstring}'.rstrip(), 67
            ).split(os.linesep)
        ))

        # Evaluate if the the prop can be a component
        if is_component_aspect(type_info):
            aspect_args.append('children=True')

        if docstring:
            aspect_args.append(
                f'docstring="{docstring.lstrip().replace(os.linesep, "")}"'
            )

        aspect = f'{name} = Aspect({", ".join(aspect_args)})'
        if len(aspect) > 75:
            aspect += '  # noqa: E501'
        aspects.append(aspect)

    if aspects:
        aspects.append('')

    aspects_optional.append('identity: str = None')

    # Enable example in docstring if they are after
    # @example with 4 spaces per lines.
    example = re.search(r'(?<=@example\n)(\s\s\s\s.*)+(?!@\w+)', description)
    if example:
        example = example.group()
        desc = description.replace(example, '').replace('@example', '').strip()
        example = f'\n\n:Example:\n\n.. code-block:: python3\n{example}'
    else:
        desc = description or ''
        example = ''

    component_string = replace_all(
        TEMPLATE,
        name=display_name,
        docstring=textwrap.indent(desc + example, '    '),
        aspects='\n'.join('    ' + x if x else x for x in aspects),
        aspects_init='\n            '.join(
            aspects_required + aspects_optional
        ),
        init_docstring='\n        '.join(init_docstring),
    ) + '\n'

    with open(output_path, 'w') as f:
        f.write(component_string)

    print(f'Generated {output_path}')


def generate_imports(output_path, components):
    with open(os.path.join(output_path, '_imports_.py'), 'w') as f:
        imports_string = '{}\n\n{}'.format(
            '\n'.join(
                'from ._{} import {}'.format(stringcase.snakecase(x), x)
                for x in components
            ),
            '__all__ = [\n{}\n]'.format(
                ',\n'.join('    "{}"'.format(x) for x in components))
        )
        imports_string += '\n'

        f.write(imports_string)


async def generate_meta(source_dir: str) -> dict:
    cmd = shlex.split(
        f'node {meta_path} {source_dir}',
        posix=sys.platform != 'win32'
    )
    proc = await asyncio.create_subprocess_shell(
        ' '.join(cmd),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    out, err = await proc.communicate()
    if err:
        print(err, file=sys.stderr)
        sys.exit(1)
    else:
        return json.loads(out.decode())


async def generate_components(metadata, output_path, executor):
    futures = []
    names = []

    for data in metadata.values():
        name = data['displayName']
        names.append(name)
        futures.append(
            executor.execute(
                generate_component,
                name,
                data['description'],
                data['props'],
                os.path.join(output_path, f'_{stringcase.snakecase(name)}.py')
            )
        )

    futures.append(executor.execute(generate_imports, output_path, names))

    await asyncio.gather(*futures)



import os

from dazzler.system import (
    Package as _Package,
    assets_to_requirements as _ass_to_req,
    Requirement as _Requirement
)

from dazzler._assets import (
    assets as _assets,
    assets_dev as _dev,
    assets_dist_path as _dist_path,
    assets_dev_path as _dev_path,
    vendors_path as _vendors
)


_name = 'renderer'
_package_name = f'dazzler_{_name}'

_vendors_requirements = [
    # React
    _Requirement(
        internal=os.path.join(_vendors, 'react-16-12-0.production.min.js'),
        dev=os.path.join(_vendors, 'react-16-12-0.development.js'),
        external='https://unpkg.com/react@16.12.0/umd/react.production.min.js',
    ),
    # React-dom
    _Requirement(
        internal=os.path.join(_vendors, 'react-dom-16-12-0.production.min.js'),
        dev=os.path.join(_vendors, 'react-dom-16-12-0.development.js'),
        external='https://unpkg.com/react-dom@16.12.0/umd/react-dom.production.min.js',  # noqa: E501
    ),
    # Normalize
    _Requirement(
        internal=os.path.join(_vendors, 'normalize-8-0-1.min.css'),
        external='https://cdnjs.cloudflare.com/ajax/libs/normalize/8.0.1/normalize.min.css'  # noqa: E501
    )
]


package = _Package(
    _package_name,
    requirements=_vendors_requirements + _ass_to_req(
        _dist_path, _assets.get('commons', {}),
        dev_data=_dev.get('commons', {}),
        dev_path=_dev_path,
        package_name=_package_name
    ) + _ass_to_req(
        _dist_path, _assets.get(_name, {}),
        dev_data=_dev.get(_name, {}),
        dev_path=_dev_path,
        package_name=_package_name,
    )
)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class CodeBlock(Component):
    """
    A block of code to highlight.

    Wrapper of ``prism-react-renderer``.

    :CSS:

         ``dazzler-markdown-code-block``
    """
    _aspect_keys = []
    _children = []
    value = Aspect(required=True, children=True, docstring="The code to render")  # noqa: E501
    language = Aspect(required=True, docstring="Language to render")
    style = Aspect()
    class_name = Aspect()

    def __init__(
            self,
            value: typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]],  # noqa: E501
            language: str,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            identity: str = None
    ):
        """
        :param value: The code to render
        :param language: Language to render
        :param style:
        :param class_name:
        """
        Component.__init__(self, locals(), identity)
